///////////////////////////////////////////////////////////////////////////
// debug.c
//
// debug code


#ifndef _MAIN_H
#include "main.h"
#endif

#ifndef __DEBUGMGR_H__
#include "debugmgr.h"
#endif

#ifndef _INC_STDARG
#include "stdarg.h"
#endif



///////////////////////////////////////////////////////////////////////////
// Assert

#ifdef GLOBALS_ALWAYS_SAFE
static Boolean s_fInDbgAssert = fFalse;
#else
#define AssertDBName		"HandyShopperAssertLock"
#define AssertType			'dbgT'
#define AssertCreator		'dbgC'
#endif

Boolean FInAssert(void);
void SetFInAssert(Boolean fInAssert);


Boolean FInAssert(void)
{
#ifdef GLOBALS_ALWAYS_SAFE
	return s_fInDbgAssert;
#else
	Boolean fInAssert;

	DbgMsg((10, "FInAssert()..."));
	fInAssert = DmFindDatabase(0, AssertDBName);
	DbgMsg((10, "FInAssert() returns %d", fInAssert));
	return fInAssert;
#endif
}


void SetFInAssert(Boolean fInAssert)
{
#ifdef GLOBALS_ALWAYS_SAFE
	s_fInDbgAssert = fInAssert;
#else
	DmOpenRef pdb;
	LocalID id;

	DbgMsg((10, "SetFInAssert(%d)", fInAssert));

	if (fInAssert)
		{
		pdb = DmCreateDatabase(0, AssertDBName, AssertCreator, AssertType, fFalse);
		if (pdb)
			DmCloseDatabase(pdb);
		else
			DbgMsg((50, "unable to create assert database!\nerror = %d",
					DmGetLastErr()));
		}
	else
		{
		id = FInAssert();
		if (id)
			DmDeleteDatabase(0, id);
		else
			DbgMsg((20, "wasn't in assert..."));
		}
#endif
}


Boolean DbgAssert(const CharPtr pszFile, UInt iLine, const CharPtr pszFmt, ...)
{
	va_list			arg;
	char			sz[256];
	CharPtr			pch = sz;
	Boolean			fRet = fTrue; // set to false to prevent this assert from firing
	int				ival;
#if 0
	char *pchEnd;
#endif

	pch += StrPrintF(pch, "%s, li %d", pszFile, iLine);

	if (pszFmt)
		{
		*pch++ = ':';
		*pch++ = ' ';
		*pch++ = '"';

		va_start(arg, pszFmt);
		pch += StrVPrintF(pch, pszFmt, arg);
		va_end(arg);
		// Remove trailing newlines...
		while (*(pch-1) == '\n')
			--pch;

		*pch++ = '"';
#if 0
		*pch++ = '\r';
		*pch++ = '\n';
#endif
		}
	else
		*pch++ = ' ';


	if (FInAssert())
		{
		*pch = 0;
		// recursive assert
		DbgMsg((20, "recursive assert!"));
		SndPlaySystemSound(sndError);SndPlaySystemSound(sndError);
		DbgMsg((50, sz));
		return fRet;
		}

	SetFInAssert(fTrue);

#if 0
	// add stack context info
	pch += wsprintfA(pch, "\nContext:%s", StackTrace::DumpToStr(fFalse));

	*pch = '\0';
	pchEnd = pch;

	Trace("%s: %s", c_szAssertTitle, (LPSTR)sz);

	// add some help about using this assert
	const char szHelp[] = "\n\nShift-Ignore: Copies assert text to the clipboard."
						  "\nCtrl-Ignore: Ignores assert for this session.\n";
	lstrcpyA(pch, szHelp);				// don't need lstrcpyn (dest is 2048 bytes)
#endif

	*pch = 0;
	ival = FrmCustomAlert(666, sz, "", "");

	switch (ival)
		{
		case 1:
			// Ignore All - return fFalse so this assert won't fire anymore
			fRet = fFalse;
			break;
		case 2:
			// Debug - break into debugger via breakpoint instruction
			DbgBreak();
			break;
		case 0:
			// Ignore - just continue
			break;
		}

	SetFInAssert(fFalse);
	return fRet;
}


static int CchToDraw(CharPtr pch, RectanglePtr prc)
{
	int				i;
	int				cch;

	cch = FldWordWrap(pch, prc->extent.x - 4);
	for (i = 0; i < cch; ++i)
		if (pch[i] == '\n')
			{
			cch = i;
			break;
			}
	return cch;
}


void _DbgMsg(int tenths, const CharPtr pszFmt, ...)
{
	va_list			arg;
	char			sz[256];
	CharPtr			pch = sz;
	WinHandle		hwin;
	RectangleType	rc;
	Err				err;
	FontID			font;
	int				y;
	int				cch;
	WinHandle		hwinDraw;
	WinHandle		hwinActive;

	va_start(arg, pszFmt);
	pch += StrVPrintF(pch, pszFmt, arg);
	va_end(arg);

	*pch = 0;

	font = FntSetFont(boldFont);

	hwinDraw = WinGetDrawWindow();
	hwinActive = WinGetActiveWindow();

	WinSetDrawWindow(WinGetDisplayWindow());

	rc.topLeft.x = 0;
	rc.topLeft.y = 0;
	rc.extent.x = 160;
	rc.extent.y = 160;
	hwin = WinSaveBits(&rc, &err);

	rc.topLeft.x = 4;
	rc.topLeft.y = 4;
	rc.extent.x = 152;

	// calc # of lines
	y = 0;
	pch = sz;
	while(*pch)
		{
		cch = CchToDraw(pch, &rc);
		y++;
		pch += cch;
		while (*pch == ' ' || *pch == '\n')
			pch++;
		}

	rc.extent.y = y * FntLineHeight() + 4;
	WinEraseRectangle(&rc, 0);
	WinDrawGrayRectangleFrame(simpleFrame, &rc);

	y = rc.topLeft.y + 2;
	pch = sz;
	while(*pch)
		{
		cch = CchToDraw(pch, &rc);
		WinDrawChars(pch, cch, rc.topLeft.x + 2, y);
		y += FntLineHeight();
		pch += cch;
		while (*pch == ' ' || *pch == '\n')
			pch++;
		}

	FntSetFont(font);

	SysTaskDelay((tenths * sysTicksPerSecond) / 10);

	if (hwin)
		WinRestoreBits(hwin, 0, 0);

	WinSetActiveWindow(hwinActive);
	WinSetDrawWindow(hwinDraw);
}



///////////////////////////////////////////////////////////////////////////
// Memory tracking

//#undef MemHandleNew
//#undef MemHandleFree
#undef MemPtrNew
#undef MemPtrFree
#undef MemPtrSize
#undef MemChunkFree


typedef struct tagMemTrack {
	struct tagMemTrack		*pNext;
	ULong					size;
	CharPtr					pszFile;
	ULong					ulLine;
	ULong					prologue;
	} MemTrack, *MemTrackPtr;

#define MEM_FILL_NEW		0xCC
#define MEM_FILL_FREE		0xDD
#define MEM_PRO				0xDEADBEEF
#define MEM_EPI				0xDE
#define MEM_OFS				sizeof(MemTrack)
#define MEM_EXTRA			MEM_OFS + 2



static ULong s_cAllocatedBytes = 0;
static ULong s_cAllocations = 0;
static MemTrackPtr s_pHead = NULL;

static ULong s_cBusy = 0;


/*

 NOTE:  the memory tracking implemented here of course does not cover things
		like having a Field allocate memory and then forgetting to call 
		FldFreeMemory.

 ALSO:  handle tracking is not implemented because i'm not really allocating 
		handles and there are some stickier issues for tracking them 
		successfully.

*/


#if 0
VoidHand __MemHandleNew(ULong size _MEMTRACKARGS)
{
	VoidHand	h;

	h = MemHandleNew(size);
	return h;
}


Err __MemHandleFree(VoidHand h)
{
	return MemHandleFree(h);
}
#endif


VoidPtr __MemPtrNew(ULong size _MEMTRACKARGS)
{
	VoidPtr		p;

	Assert(!s_cBusy);
	s_cBusy++;

	p = MemPtrNew(size + MEM_EXTRA);
	if (p)
		{
		MemTrackPtr pTrack = (MemTrackPtr)p;

		p += MEM_OFS;
		MemSet(p + size, 2, MEM_EPI);

		pTrack->pNext = s_pHead;
		pTrack->pszFile = pszFile;
		pTrack->ulLine = ulLine;
		pTrack->size = size;
		pTrack->prologue = MEM_PRO;

		s_pHead = pTrack;

		MemSet(p, size, MEM_FILL_NEW);

		s_cAllocatedBytes += size;
		s_cAllocations++;

		Assert(((ULong)p) - ((ULong)pTrack) == MEM_OFS);
//Panic3("pTrack = %ld, p = %ld, size = %ld", pTrack, p, size);
		}
	s_cBusy--;
	return p;
}


void __UnlinkPtr(VoidPtr p)
{
	MemTrackPtr *pp;
	MemTrackPtr pTrack = (MemTrackPtr)(p - MEM_OFS);

	for (pp = &s_pHead; *pp && *pp != pTrack; pp = &((*pp)->pNext));
	AssertSz(*pp == pTrack, "unlinking a pointer that wasn't being tracked");
	if (*pp)
		*pp = (*pp)->pNext;
}


static void VerifyBlock(MemTrackPtr pTrack)
{
	ULong size = MemPtrSize(pTrack) - MEM_EXTRA;
	UInt c;
	BytePtr pb;

	AssertFL0(pTrack->pszFile, pTrack->ulLine, pTrack->prologue == MEM_PRO,
			  "memory prerun");
	AssertFL2(pTrack->pszFile, pTrack->ulLine, pTrack->size <= size,
			  "expected size %ld, actual size %ld", pTrack->size, size);

	for (c = 2, pb = ((CharPtr)pTrack) + sizeof(*pTrack) + pTrack->size; *pb == MEM_EPI && c; ++pb, c--);
	AssertFL0(pTrack->pszFile, pTrack->ulLine, !c, "memory overrun");
}


Err __MemChunkFree(VoidPtr p)
{
	Err err = 0;
	MemTrackPtr pTrack;

	Assert(!s_cBusy);
	s_cBusy++;

	Assert(p);
	if (p)
		{
		pTrack = (MemTrackPtr)((CharPtr)p - MEM_OFS);

//Panic3("pTrack = %ld, p = %ld, size = %ld", pTrack, p, size);
		VerifyBlock(pTrack);

		// break the link
		__UnlinkPtr(p);
		s_cAllocatedBytes -= pTrack->size;
		s_cAllocations--;

		// fill value
		MemSet(p, pTrack->size, MEM_FILL_FREE);

		// free the pointer
		err = MemChunkFree(pTrack);
		}
	s_cBusy--;
	return err;
}


void __ListLeaks(void)
{
	MemTrackPtr pTrack;
	Char szBytes[10];
	Char szLine[10];
	ULong cLeaks = 0;

	Assert1(!s_cAllocations, "%ld leaks!", s_cAllocations);
	if (!s_cAllocations && s_pHead)
		PanicSz("allocation count does not match allocation list");

	for (pTrack = s_pHead; pTrack; pTrack = pTrack->pNext, cLeaks++)
		{
		VerifyBlock(pTrack);

		StrIToA(szBytes, pTrack->size);
		StrIToA(szLine, pTrack->ulLine);
		FrmCustomAlert(667, szBytes, pTrack->pszFile, szLine);
		}

	Assert2(cLeaks == s_cAllocations,
			"allocation count = %ld, allocation list contains %ld elements!",
			s_cAllocations, cLeaks);
}


void __ListTrackInfo(void)
{
	Char szBytes[10];
	Char szCount[10];
	Char szOverhead[10];

	StrIToA(szBytes, s_cAllocatedBytes);
	StrIToA(szCount, s_cAllocations);
	StrIToA(szOverhead, s_cAllocations * MEM_EXTRA);

	FrmCustomAlert(668, szCount, szBytes, szOverhead);
}


