#ifdef _DEBUG_H
#error _DEBUG_H already defined!
#else
#define _DEBUG_H
#endif

// globals aren't always safe with GCC.  however, compiling without this
// defined disables the "Ignore All" feature on asserts (without removing the
// button, though).  it is such a useful feature that we'll always compile
// with this defined, except when tracking down strange crashes.
#define GLOBALS_ALWAYS_SAFE

extern Boolean AssertFormHandleEvent(EventPtr pEvent);



#define Panic()								Assert0(0, "Panic")
#define PanicSz(szFmt)						Assert0(0, szFmt)
#define Panic0(szFmt)						Assert0(0, szFmt)
#define Panic1(szFmt, p1)					Assert1(0, szFmt, p1)
#define Panic2(szFmt, p1, p2)				Assert2(0, szFmt, p1, p2)
#define Panic3(szFmt, p1, p2, p3)			Assert3(0, szFmt, p1, p2, p3)
#define Panic4(szFmt, p1, p2, p3, p4)		Assert4(0, szFmt, p1, p2, p3, p4)
#define Panic5(szFmt, p1, p2, p3, p4, p5)	Assert5(0, szFmt, p1, p2, p3, p4, p5)
#define AssertSz(f, sz)						Assert0(f, sz)



#ifdef DEBUG
	// Debug

	#define IfDebug(expr) expr

	// override memory allocation routines
	#define _MEMTRACKARGS		, CharPtr pszFile, ULong ulLine
	#define _MEMTRACK			, __FILE__, __LINE__

	void __ListLeaks(void);
	void __ListTrackInfo(void);

//	VoidHand __MemHandleNew(ULong size _MEMTRACKARGS);
//	Err __MemHandleFree(VoidHand h);
//	Void __UnlinkHandle(VoidHand h);

	VoidPtr __MemPtrNew(ULong size _MEMTRACKARGS);
	Err __MemChunkFree(VoidPtr p);
	void __UnlinkPtr(VoidPtr p);

//	#define MemHandleNew __MemHandleNew
//	#define MemHandleFree __MemHandleFree

	#undef MemPtrFree
	#define MemPtrNew __MemPtrNew
	#define MemPtrFree __MemChunkFree
	#define MemChunkFree __MemChunkFree

	#define MemChunkNew __DoNotUse_MemChunkNew__
	#define MemPtrResize __DoNotUse_MemPtrResize__
	#define MemPtrSize __DoNotUse_MemPtrSize__
	#define MemPtrRecoverHandle __DoNotUse_MemPtrRecoverHandle__


	// asserts ------------------------------------------------------------
	Boolean DbgAssert(const CharPtr pszFile, UInt iLine, const CharPtr pszFmt, ...);
	void _DbgMsg(int tenths, const CharPtr pszFmt, ...);
	#define Verify(expr) Assert(expr)

#ifdef GLOBALS_ALWAYS_SAFE
	#define PreAssert(f)	do { static Boolean fDoIt = fTrue; if (!(f) && fDoIt) fDoIt =
	#define PostAssert		} while (0)
#else
	#define PreAssert(f)	do { if (!(f))
	#define PostAssert		} while (0)
#endif

	#define DbgMsg(arglist) _DbgMsg arglist

	#define Assert(f) \
		PreAssert(f) \
		DbgAssert(__FILE__, __LINE__, "%s", #f); \
		PostAssert
	#define Assert0(f, szFmt) \
		PreAssert(f) \
		DbgAssert(__FILE__, __LINE__, szFmt); \
		PostAssert
	#define Assert1(f, szFmt, p1) \
		PreAssert(f) \
		DbgAssert(__FILE__, __LINE__, szFmt, p1); \
		PostAssert
	#define Assert2(f, szFmt, p1, p2) \
		PreAssert(f) \
		DbgAssert(__FILE__, __LINE__, szFmt, p1, p2); \
		PostAssert
	#define Assert3(f, szFmt, p1, p2, p3) \
		PreAssert(f) \
		DbgAssert(__FILE__, __LINE__, szFmt, p1, p2, p3); \
		PostAssert
	#define Assert4(f, szFmt, p1, p2, p3, p4) \
		PreAssert(f) \
		DbgAssert(__FILE__, __LINE__, szFmt, p1, p2, p3, p4); \
		PostAssert
	#define AssertFL0(szFile, iLine, f, szFmt) \
		PreAssert(f) \
		DbgAssert(szFile, iLine, szFmt); \
		PostAssert
	#define AssertFL2(szFile, iLine, f, szFmt, p1, p2) \
		PreAssert(f) \
		DbgAssert(szFile, iLine, szFmt, p1, p2); \
		PostAssert

#if 1
	//$ todo: (chrisant) not sure good way to do leave tracing yet (use C++ mode?)
	#define CONTEXT_(sz) \
		{ \
		CharPtr __szFunc = sz; \
		if (g_fEnterLeaveTraces) \
			PanicSz("enter "#sz);
	#define LEAVE() \
		if (g_fEnterLeaveTraces) \
			{ \
			Char sz[64]; \
			StrCopy(sz, "leave "); \
			StrCat(sz, __szFunc); \
			PanicSz(sz); \
			} \
		}
#else
	#define CONTEXT_(sz)
	#define LEAVE()
#endif

#else
	// Retail

	#define IfDebug(expr)

	#define _MEMTRACK
	#define __ListLeaks()

	// asserts ------------------------------------------------------------
	#define Verify(expr) expr
	#define Assert(fExpr)
	#define Assert0(fExpr, sz)
	#define Assert1(fExpr, sz, arg1)
	#define Assert2(fExpr, sz, arg1, arg2)
	#define Assert3(fExpr, sz, arg1, arg2, arg3)
	#define Assert4(fExpr, sz, arg1, arg2, arg3, arg4)

	#define DbgMsg(arglist)

	#define CONTEXT_(sz)
	#define LEAVE()

#endif

