///////////////////////////////////////////////////////////////////////////
// globals.h

#ifdef _GLOBALS_H
#error _GLOBALS_H already defined!
#else
#define _GLOBALS_H
#endif


#ifndef _MAIN_H
#include "main.h"
#endif



#ifdef DECLARE_GLOBALS
#define GLOBAL(typ, var) typ var;
#define GLOBALx(typ, var)
#define GLOBAL_(typ, var, val) typ var = val;

#else
#define GLOBAL(typ, var) extern typ var;
#define GLOBALx(typ, var) extern typ var;
#define GLOBAL_(typ, var, val) extern typ var;

#endif



#ifdef DEBUG
GLOBAL_(Boolean, g_fEnterLeaveTraces, fFalse)
#endif



///////////////////////////////////////////////////////////////////////////
// Types

#define MAX_COLUMNS			7
#define MAX_ROWS			10
#define MAX_STORES			15

#define COL_NEED			0
#define COL_NAME			1
#define COL_QUANTITY		2
#define COL_COUPON			3
#define COL_AISLE			4
#define COL_CATEGORY		5
#define COL_PRICE			6

#define cchMaxText			(254)
#define cchMaxPriceText		(12)
#define cchMaxStoreName		(32)
#define cchMaxLookup		(10)


// P11. An application info structure is intended to be stored in the
// database's app info block and to contain data relevant to the database.
// For now we store just category information.  Category information is
// always placed in the beginning where the category.c code expects it.
typedef struct
	{
	// categories
	UInt	renamedCategories;		// bitfield of categories with a changed name
	char	categoryLabels[dmRecNumCategories][dmCategoryLength];
	Byte	categoryUniqIDs[dmRecNumCategories];
	Byte	lastUniqID;			// Uniq IDs for categories generated by the device
									// are between 0 - 127.  Those from the PC are 128 - 255.
	//Byte	reserved1;			// from the compiler word aligning things
	//Word	reserved2;

	// stores
	UInt				cStores;
	UInt				bitStores;				// mask of used store bits
	UInt				rgbitStores[MAX_STORES];
	UInt				rgcItemsInStore[MAX_STORES+1];	// [0] is "All Stores" (total) count
	Char				rgszStoreNames[MAX_STORES][cchMaxStoreName+1];	// inefficient but ever-so-much simpler to code
	} ThisAppInfoType;

typedef ThisAppInfoType *ThisAppInfoPtr;		// P11.
typedef CharPtr DBRecordTypePtr;				// P14.


// P15. The data saved between application executions.  This data
// is saved as a single resource in the System Preferences.
#define prefOneTime			0x0001
#define prefUseDecimals		0x0002
#define prefAutoQuantity	0x0004
typedef struct
	{
	UInt				irecTopVisible;

	UInt				flags;

	UInt				store;
	UInt				view;
	UInt				columns;
	UInt				catCurrent;
	} PreferenceType;


enum { viewAll, viewNeed, viewCoupon, viewMAX };

enum {
	colNeed				= 0x01,
	colName				= 0x02,
	colQuantity			= 0x04,
	colCoupon			= 0x08,
	colAisle			= 0x10,
	colCategory			= 0x20,
	colPrice			= 0x40,
	};


typedef struct
	{
	UInt				bit;
	UInt				cItems;
	CharPtr				pszName;
	} Store;


#define flagNeed		0x01
#define flagCoupon		0x02
#define flagPrivate		0x80			// private (home-brewed so we can keep store bits in sync)

#define storeOnce		0x8000

typedef struct
	{
	UChar				flags;
	UChar				cchName;		// length of item description, excluding null terminator
	Word				store;			// 16 bits mapping to stores
	UInt				uQuantity;
	ULong				ulPrice;
	UChar				aisle;
	Char				szName[0];		// the item description
	} Item, *ItemPtr;



///////////////////////////////////////////////////////////////////////////
// Defines

// Application database management defines.
#define AppType				'iShp'				// P5. type for application.  must be 4 chars, mixed case.
#define DBType				'Data'				// P5. type for application database.  must be 4 chars, mixed case.
#define DBName				"HandyShopperDB"	// P5. name for application database.  up to 31 characters.
#define PrefsVersionNum		21					// preferences format version
#define noRecordSelected	((UInt)-1)			// P5. index for no current record.
#define noRecordSelectedID	((UInt)-1)			// P15. id for no unique id

#define firstRecord			0					// index of first data record

#define updateCategoryChanged		0x01		// P11. Indicates a category changed
#define updateStoreChanged			0x02		// indicates store bits changed


// MemoPad stuff (for Export to MemoPad command)
#define memoPadAppType		'memo'
#define memoPadDBType		'DATA'
#define memoPadDBName		"MemoDB"



///////////////////////////////////////////////////////////////////////////
// Constants



///////////////////////////////////////////////////////////////////////////
// Variables

GLOBAL(SystemPreferencesPtr, g_pSysPrefs)
GLOBAL(Char, g_chSep)
GLOBAL(Char, g_chDecimal)

GLOBAL(DmOpenRef, g_pdb)						// P5. handle for database
GLOBAL(Char, szCategoryName[dmCategoryLength])	// P11. Storage for the name
GLOBAL(Word, catRecord)							// P11. edited record's category (never All)
GLOBAL(Word, catNew)							// P11. category in details form (never All)
GLOBAL(Boolean, g_fNewCategoryEdited)			// P11. true if name edited

GLOBAL(UInt, g_cVisibleItems)					// number of visible items in table
GLOBAL(Int, g_cLinesRemaining)					// remaining (unused) rows in table

GLOBAL_(MenuBarPtr, pmenuCurrent, NULL)

GLOBAL(UChar, g_cxMaxStorePixels)



// P15. The following global variable are saved to a state file.
GLOBAL_(UInt, recCurrent, noRecordSelected)		// P5. index of current database record
GLOBAL_(UInt, catCurrent, dmAllCategories)		// P11. category filter
GLOBAL_(UInt, irecTopVisible, firstRecord)		// P12. top record in main table
GLOBAL_(Boolean, g_fOnceDefault, fFalse)
GLOBAL_(Boolean, g_fDecimals, fTrue)			// show prices with 2 decimal places (fFalse = round to whole number)
GLOBAL_(Boolean, g_fAutoQuantity, fTrue)		// automatically reset quantities
GLOBAL_(Boolean, g_fEditingPrice, fFalse)
GLOBAL_(Word, g_store, 0)
GLOBAL_(UInt, g_view, viewNeed)
GLOBAL_(UInt, g_columns, colNeed|colName|colQuantity|colPrice)
GLOBAL_(UInt, g_cStores, 0)

// [0].bit is mask of all used bits for stores, [0].pszName is "All Stores"
GLOBAL_(Store, g_rgStores[MAX_STORES+1], {{0}})



