///////////////////////////////////////////////////////////////////////////
// options.c
//
// the Options form

#ifndef _MAIN_H
#include "main.h"
#endif

#ifndef _OPTIONS_H
#include "options.h"
#endif

#include "res.h"



///////////////////////////////////////////////////////////////////////////
// Options Form

static const struct {
	UInt idc;
	UChar col;
	} c_rgcols[] = { {IDC_OPTIONS_QUANTITY, colQuantity},
					 {IDC_OPTIONS_COUPON, colCoupon},
					 {IDC_OPTIONS_AISLE, colAisle},
					 {IDC_OPTIONS_CATEGORY, colCategory},
					 {IDC_OPTIONS_PRICE, colPrice},
					 };


static Boolean Apply(void)
{
	Int ii;
	UChar columns = g_columns;
	Boolean f;
	Boolean fResetForm = fFalse;

	g_columns = colNeed | colName;
	for (ii = DimensionOf(c_rgcols); ii--;)
		if (CtlGetValue(GetObjectPtr(c_rgcols[ii].idc)))
			g_columns |= c_rgcols[ii].col;
	if (g_columns != columns)
		fResetForm = fTrue;

	g_fOnceDefault = CtlGetValue(GetObjectPtr(IDC_OPTIONS_ONCE));
	g_fAutoQuantity = CtlGetValue(GetObjectPtr(IDC_OPTIONS_AUTOQUANTITY));

	f = !CtlGetValue(GetObjectPtr(IDC_OPTIONS_ROUNDING));
	if (g_fDecimals != f)
		fResetForm = fTrue;
	g_fDecimals = f;

	return fResetForm;
}


static void InitForm(void)
{
	Int ii;

	Assert((g_columns & (colNeed|colName)) == (colNeed|colName));

	for (ii = DimensionOf(c_rgcols); ii--;)
		CtlSetValue(GetObjectPtr(c_rgcols[ii].idc), F(g_columns & c_rgcols[ii].col));

	CtlSetValue(GetObjectPtr(IDC_OPTIONS_ONCE), g_fOnceDefault);
	CtlSetValue(GetObjectPtr(IDC_OPTIONS_ROUNDING), !g_fDecimals);
	CtlSetValue(GetObjectPtr(IDC_OPTIONS_AUTOQUANTITY), g_fAutoQuantity);
}


Boolean OptionsFormHandleEvent(EventPtr pEvent)
{
	Boolean		fHandled = fTrue;

	CALLBACK_PROLOGUE

	switch (pEvent->eType)
		{
		case ctlSelectEvent:
			// A control button was pressed and released.
			switch (pEvent->data.ctlSelect.controlID)
				{
				case IDC_FORM_OK:
					if (Apply())
						{
						FrmCloseAllForms();
						FrmGotoForm(FRM_MAIN);
						break;
						}
					// else fall thru
				case IDC_FORM_CANCEL:
					FrmReturnToForm(0);
					RestoreInsPt();
					break;

				default:
					fHandled = fFalse;
					break;
				}
			break;

		case frmOpenEvent:
			InitForm();
			FrmDrawForm(FrmGetActiveForm());
			break;

		case frmCloseEvent:
			// this does NOT get sent by FrmReturnToForm.  must delete the
			// form when closed other than by FrmReturnToForm.
			FrmDeleteForm(FrmGetActiveForm());
			break;

		default:
			fHandled = fFalse;
		}

	CALLBACK_EPILOGUE
	return fHandled;
}



