///////////////////////////////////////////////////////////////////////////
// shop.c
//
// Shopping List Manager
//
// Christopher Antos
// chrisant@premier1.net


#ifndef _SHOP_H
#include "shop.h"
#endif

#ifndef _OPTIONS_H
#include "options.h"
#endif

#ifndef _DETAILS_H
#include "details.h"
#endif

#ifndef _STORES_H
#include "stores.h"
#endif

#ifndef __MEMORYMGR_H__
#include "memorymgr.h"
#endif

#include "res.h"


//$ todo: (chrisant) if you accidentally check off a "one time only" item, you lose it!  what are some ways around this?

//$ todo: (chrisant) deadline date - when to purchase (and snoozeable reminder?)
//$ todo: (chrisant) reminder?  per item or per category?
//$ todo: (chrisant) Impulse Date (date you see something, so you can save money avoiding the impulses where you later kick yourself)
//$ todo: (chrisant) option to record date of purchase
//$ todo: (chrisant) recurring items, don't show up until the recurrence date; if deadline dates are implemented, need to say when the deadline date is, too.
//$ todo: (chrisant) import from another database?  cut/copy/paste items from one database to another?
//$ todo: (chrisant) import/export via memopad?
//$ todo: (chrisant) scrolling table for stores in Details form

//$ todo: (chrisant) live sorting?

//$ todo: (chrisant) multiple databases?
//$ todo: (chrisant) reminder per database?



/*

v1.01 ---------------------------------------------------------------------
	- FIX: Lookup was finding the last possible match instead of the first 
	  possible match.  [Fixed by adding "break;" when found a match (duh)]
	- FIX: On Palm3 only, Details form sometimes caused an error requiring a 
	  Reset.  [Fixed by adding CALLBACK_PROLOGUE/CALLBACK_EPILOGUE to 
	  DetailsFormHandleEvent]
	- FIX: The hardware Down button caused a error requiring reset, if you had 
	  around 20+ items in the list.  IncrementalLoadTableBackwards had a loop
	  with no upper bound, which caused an invalid parameter to be passed to
	  the OS.  [Fixed by adding the correct upper bound]
	- FIX: Using Cut/Copy/Paste (or any menu command) in any form except 
	  FRM_MAIN would cause an error requiring a reset (was always trying to 
	  get a pointer to the table control).  [Fixed by only getting a pointer 
	  to the table control when the active form is the FRM_MAIN form]
	- FIX: Prices did not use the system preferences for decimals and 
	  separator.  [Fixed by checking the system prefs]
	- FIX: Holding the hardware Down button made the table keep trying to page 
	  down even when it was already at the bottom, which made the table 
	  flicker.  [Fixed by checking if we're already at the bottom, before 
	  trying to page down]

v1.02 ---------------------------------------------------------------------
	- FIX: Checking off an item with a blank description also tried to check 
	  off the following item, and would crash if there were no more items.
	- FIX: Trying to install a HandyShopperDB.PDB file after a hard reset 
	  caused the PalmPilot to crash.  This is because when handling the 
	  sysAppLaunchCmdInitDatabase launch code, we do not have access to our 
	  global variables (not even if we use CALLBACK_PROLOGUE).  [Fixed by 
	  removing all global variable usage from that code path]
	- Display error and refuse to start on OS 1.x devices.

v1.03 ---------------------------------------------------------------------
	- FIX: The system Find command was finding items marked private even when 
	  the Pilot was set to hide private records.  [Fixed by checking the 
	  item's private bit before calling it a match]
	- Up/Down arrows on the screen scrolled by one item at a time.  Now they 
	  scroll a page at a time, same as the hard Up/Down buttons.
	- FIX: the hard Down button could crash if there were no items in the 
	  list.
	- Price column is right justified.
	- German translation by Ronald Hoellwarth.  (Someone else also offered to
	  translate it to German, but I lost track of their email address).
	- FIX: If the last item in the table was more than one line, sometimes 
	  only the first line was shown.  [Fixed by checking if the ideal height 
	  of the last item is greater than the height it occupies -- not the best 
	  fix, but HS2.0 won't even use a Table control (i'm doing it myself) so 
	  it's not worth fixing for HS1.x]

*/



///////////////////////////////////////////////////////////////////////////
// Functions

static void TblCallback_DrawPrice(VoidPtr ptbl, Word row, Word column, RectanglePtr prc);


static Err InitAppInfo(DmOpenRef pdb, Boolean fGlobalsOk)
{
	UInt		cardNo;
	VoidHand	h;
	LocalID		idDB;
	LocalID		idAppInfo;
	AppInfoPtr	pAppInfo;

	UInt		ui;
	Byte		b;

	//$ review: (chrisant) for some reason, CALLBACK_PROLOGUE does not help
	// here if we are inside a sysAppLaunchCmdInitDatabase launch.

#ifdef DEBUG
	if (!pdb)
		{
		if (fGlobalsOk)
			PanicSz("InitAppInfo got a NULL pdb!");
		else
			DbgMsg((50, "InitAppInfo got a NULL pdb!"));
		}
#endif

	// We have a DmOpenRef and we want the database's app info block ID.
	// Get the database's dbID and cardNo and then use them to get the appInfoID.
	if (DmOpenDatabaseInfo(pdb, &idDB, NULL, NULL, &cardNo, NULL))
		return dmErrInvalidParam;
	if (DmDatabaseInfo(cardNo, idDB, NULL, NULL, NULL, NULL, NULL,
					   NULL, NULL, &idAppInfo, NULL, NULL, NULL))
		return dmErrInvalidParam;


	// If no appInfoID exists then we must create a new one.
	if (idAppInfo == NULL)
		{
		DbgMsg((20, "creating new AppInfo block"));
		h = DmNewHandle(pdb, sizeof(ThisAppInfoType));
		if (!h)
			return dmErrMemError;

		idAppInfo = MemHandleToLocalID(h);
		DmSetDatabaseInfo(cardNo, idDB, NULL, NULL, NULL, NULL, NULL,
						  NULL, NULL, &idAppInfo, NULL, NULL, NULL);
		}


	// Lock the appInfoID and copy in defaults from our default structure.
	pAppInfo = MemLocalIDToLockedPtr(idAppInfo, cardNo);

	// initialize
	DmSet(pAppInfo, 0, sizeof(ThisAppInfoType), 0);

	ui = 6;
	AssertSameSize(pAppInfo->renamedCategories, ui);
	DmWrite(pAppInfo, OffsetOf(*pAppInfo, renamedCategories), &ui, sizeof(ui));

	b = 15;
	AssertSameSize(pAppInfo->lastUniqID, b);
	DmWrite(pAppInfo, OffsetOf(*pAppInfo, lastUniqID), &b, sizeof(b));

	AssertSameSize(pAppInfo->categoryUniqIDs[0], b);
	for (b = 0; b < 16; b++)
		DmWrite(pAppInfo, OffsetOf(*pAppInfo, categoryUniqIDs[0]) + b, &b, sizeof(b));

#if defined(GERMAN)
	DmWrite(pAppInfo, OffsetOf(*pAppInfo, categoryLabels[0]), "Unfiled", 7);         // $ translate: (hoellwarth) I checked my german addressbook labels ... even 3COM didn't translate unfiled
	DmWrite(pAppInfo, OffsetOf(*pAppInfo, categoryLabels[1]), "Business", 8);
	DmWrite(pAppInfo, OffsetOf(*pAppInfo, categoryLabels[2]), "Persnlich", 10);
#else
	DmWrite(pAppInfo, OffsetOf(*pAppInfo, categoryLabels[0]), "Unfiled", 7);
	DmWrite(pAppInfo, OffsetOf(*pAppInfo, categoryLabels[1]), "Business", 8);
	DmWrite(pAppInfo, OffsetOf(*pAppInfo, categoryLabels[2]), "Personal", 8);
#endif

	MemPtrUnlock(pAppInfo);


	// default to Unfiled to make the category trigger more discoverable
	if (fGlobalsOk)
		catCurrent = CategoryFind(pdb, "Unfiled");

	// we just blew away any store information, so make sure we zero the store
	// bits for all items so the store counts don't get off due to random bits
	// lying around.
#ifdef DEBUG
	if (fGlobalsOk)
		PanicSz("first run?  zapping stores");
	else
		DbgMsg((20, "first run?\nzapping stores"));
#endif
	ZapStore(pdb, ~storeOnce);
	return 0;
}


static Boolean StartApplication(void)
{
	Err				err = 0;			// P5. error code
	UInt			cardNo;				// P5. card containing the application database
	LocalID			dbID;				// P5. handle for application database
	UInt			dbAttrs;			// P5. database attributes
	UInt			mode;				// P5. permissions when opening the database
	PreferenceType	*pprefs = 0;		// P15. app state before closed last time
	LocalID			idAppInfo;			// P15.
	UInt			version;			// P17. The database's version.
	Int				ii;
	ThisAppInfoPtr	pAppInfo;
	CharPtr			psz;
	DWord			dwOSVersion;
	Word			cbPrefs;

	if (FtrGet(sysFtrCreator, sysFtrNumROMVersion, &dwOSVersion) != 0 ||
		dwOSVersion < 0x02000000)
		{
		Error(errNeedOS2);
		err = -1;
		}

#ifdef DEBUG
	for (ii = MAX_STORES; ii--;)
		Assert(g_rgStores[ii+1].pszName == 0);
	Assert(FntCharHeight() == FntLineHeight());
#endif

	g_pSysPrefs = MemPtrNew(sizeof(SystemPreferencesType) _MEMTRACK);
	CPE(g_pSysPrefs);
	PrefGetPreferences(g_pSysPrefs);

	switch (g_pSysPrefs->numberFormat)
		{
		case nfCommaPeriod:
			g_chSep = ',';
			g_chDecimal = '.';
			break;
		case nfPeriodComma:
			g_chSep = '.';
			g_chDecimal = ',';
			break;
		case nfSpaceComma:
			g_chSep = ' ';
			g_chDecimal = ',';
			break;
		case nfApostrophePeriod:
			g_chSep = '\'';
			g_chDecimal = '.';
			break;
		case nfApostropheComma:
			g_chSep = '\'';
			g_chDecimal = ',';
			break;
		default:
			Panic1("unknown numberFormat in system prefs: %d",
				   g_pSysPrefs->numberFormat);
		}

	pprefs = MemPtrNew(sizeof(*pprefs) _MEMTRACK);
	CPE(pprefs);

#if 0
	// P10. Determime if secret record should be shown.
	if (g_pSysPrefs->hideSecretRecords)
		mode = dmModeReadWrite;
	else
		mode = dmModeReadWrite | dmModeShowSecret;
#endif
	mode = dmModeReadWrite;

	// P5. Find the application's database.
	g_pdb = DmOpenDatabaseByTypeCreator(DBType, AppType, mode);
	if (!g_pdb)
		{
		// P5. The database doesn't exist, create it now.
		CE(DmCreateDatabase(0, DBName, AppType, DBType, fFalse));

		// P5. Find the application's database.
		g_pdb = DmOpenDatabaseByTypeCreator(DBType, AppType, mode);

		//$ review: (chrisant) error handling?

		// P5. Get info about the database
		CE(DmOpenDatabaseInfo(g_pdb, &dbID, NULL, NULL, &cardNo, NULL));

		// P5. Get attributes for the database
		CE(DmDatabaseInfo(0, dbID, NULL, &dbAttrs, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL));

		// P17. Set the backup flag in the attributes
		// This will ensure that the database will always get backed up
		// during HotSync, even if there is no conduit for this application.
		//$ review: (chrisant) does the backup bit stay on, or does it get cleared once it's been backed up once?
		dbAttrs = dbAttrs | dmHdrAttrBackup;

		// P5. Set the new attributes in the database
		CE(DmSetDatabaseInfo(0, dbID, NULL, &dbAttrs, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL));

		// P11. Store category info in the application's information block.
		// If error return so.
		if (InitAppInfo(g_pdb, fTrue))
			goto Error;
		}
	else
		// P17. After a hard reset, the records may be sent to the device.
		// In this case the database might not have an app info block.
		// Check for an app info block, and if there isn't one create a
		// default one.
		{
		// P17. Get info about the database
		CE(DmOpenDatabaseInfo(g_pdb, &dbID, NULL, NULL, &cardNo, NULL));

		// P17. Get attributes for the database.  This is an apropriate place
		// to get the version number of the database for applications which
		// have multiple versions of databases.  If the application detects
		// a different version of the database it may decide to not use it,
		// to translate it, or not support some new or old features.
		CE(DmDatabaseInfo(0, dbID, NULL, NULL, &version, NULL, NULL,
						  NULL, NULL, &idAppInfo, NULL, NULL, NULL));

		if (idAppInfo == NULL)
			{
			// P17. Store category info in the application's information block.
			if (InitAppInfo(g_pdb, fTrue))
				goto Error;
			}
		}

	// get application prefs
#if defined(GERMAN)
	g_rgStores[0].pszName = "Alle Geschfte";
#else
	g_rgStores[0].pszName = "All Stores";
#endif

	// initialize store info -- also serves as back door in case i have any
	// bugs in the bit allocation system.  deleting all stores will reset the
	// bit allocation mask to 0.
	g_rgStores[0].bit = 0;
	g_rgStores[0].cItems = DmNumRecordsInCategory(g_pdb, dmAllCategories);

	// get app info (dbID and CardNo have been initialized above)
	CE(DmDatabaseInfo(0, dbID, NULL, NULL, &version, NULL, NULL,
					  NULL, NULL, &idAppInfo, NULL, NULL, NULL));
	pAppInfo = MemLocalIDToLockedPtr(idAppInfo, cardNo);
	CPE(pAppInfo);
	g_cStores = pAppInfo->cStores;
	if (g_cStores)
		{
		g_rgStores[0].bit = pAppInfo->bitStores;
		g_rgStores[0].cItems = pAppInfo->rgcItemsInStore[0];
		}
	for (ii = g_cStores; ii--;)
		{
		g_rgStores[ii+1].cItems = pAppInfo->rgcItemsInStore[ii+1];
		g_rgStores[ii+1].bit = pAppInfo->rgbitStores[ii];
		if (pAppInfo->rgbitStores[ii])
			{
			Assert(StrLen(pAppInfo->rgszStoreNames[ii]) <= cchMaxStoreName);

			psz = MemPtrNew(StrLen(pAppInfo->rgszStoreNames[ii])+1 _MEMTRACK);
			CPE(psz);
			g_rgStores[ii+1].pszName = psz;
			StrCopy(psz, pAppInfo->rgszStoreNames[ii]);
			}
		}
	MemPtrUnlock(pAppInfo);

	// try current version format
	cbPrefs = sizeof(*pprefs);
	if (PrefGetAppPreferences(AppType, 1, pprefs, &cbPrefs,
							  fTrue) == PrefsVersionNum ||
			PrefGetAppPreferencesV10(AppType, PrefsVersionNum, pprefs,
									 sizeof(*pprefs)))
		{
		irecTopVisible = pprefs->irecTopVisible;
		catCurrent = pprefs->catCurrent;
		g_store = pprefs->store;
		g_view = pprefs->view;
		g_columns = pprefs->columns;

		g_fOnceDefault = pprefs->flags & prefOneTime;
		g_fDecimals = pprefs->flags & prefUseDecimals;
		g_fAutoQuantity = pprefs->flags & prefAutoQuantity;
		}
	else
		{
		// if unable to get a version of prefs we understand, then use
		// defaults (do nothing).
		}

	// P11. Get the name of the current category from the app info block.
	// The app info block was created and initialized by InitAppInfo.
	CategoryGetName(g_pdb, catCurrent, szCategoryName);

	// P17. The category may no longer exist if it was removed during a sync
	if (*szCategoryName == 0)
		catCurrent = dmAllCategories;

	irecTopVisible = firstRecord;
	recCurrent = noRecordSelected;

Error:
	if (pprefs)
		MemPtrFree(pprefs);
	if (FFailed(err))
		{
		if (g_pSysPrefs)
			{
			MemPtrFree(g_pSysPrefs);
			g_pSysPrefs = 0;
			}
		for (ii = g_cStores; ii--;)
			if (g_rgStores[ii+1].pszName)
				{
				MemPtrFree(g_rgStores[ii+1].pszName);
				g_rgStores[ii+1].pszName = 0;
				}
		}
	return FSuceeded(err);
}


static void StopApplication(void)
{
	Err				err = 0;
	PreferenceType	*pprefs;
	Int				ii;
	ThisAppInfoPtr	pAppInfo;
	UInt			cardNo;				// P5. card containing the application database
	LocalID			dbID;				// P5. handle for application database
	LocalID			idAppInfo;			// P15.
	UInt			version;			// P17. The database's version.

	// P11. Close all open forms to allow their frmCloseEvent handlers
	// to execute.  An appStopEvent doesn't send frmCloseEvents.
	// FrmCloseAllForms will send all opened forms a frmCloseEvent.
	FrmCloseAllForms();

	//$ todo: (chrisant) if anything at all changed, set the backup attribute
	// on the database.

	pprefs = MemPtrNew(sizeof(PreferenceType) _MEMTRACK);
	CPE(pprefs);

	// Write the AppInfo block (contains categories and stores)
	CE(DmOpenDatabaseInfo(g_pdb, &dbID, NULL, NULL, &cardNo, NULL));
	CE(DmDatabaseInfo(0, dbID, NULL, NULL, &version, NULL, NULL,
					  NULL, NULL, &idAppInfo, NULL, NULL, NULL));
	pAppInfo = MemLocalIDToLockedPtr(idAppInfo, cardNo);
	CPE(pAppInfo);
		AssertSameSize(pAppInfo->cStores, g_cStores);
	DmWrite(pAppInfo, OffsetOf(*pAppInfo, cStores), &g_cStores, sizeof(g_cStores));
		AssertSameSize(pAppInfo->rgcItemsInStore[0], g_rgStores[0].cItems);
	DmWrite(pAppInfo, OffsetOf(*pAppInfo, rgcItemsInStore[0]), &g_rgStores[0].cItems, sizeof(g_rgStores[0].cItems));
		AssertSameSize(pAppInfo->bitStores, g_rgStores[0].bit);
	DmWrite(pAppInfo, OffsetOf(*pAppInfo, bitStores), &g_rgStores[0].bit, sizeof(g_rgStores[0].bit));
	for (ii = g_cStores; ii--;)
		{
			AssertSameSize(pAppInfo->rgcItemsInStore[ii+1], g_rgStores[ii+1].cItems);
		DmWrite(pAppInfo, OffsetOf(*pAppInfo, rgcItemsInStore[ii+1]), &g_rgStores[ii+1].cItems, sizeof(g_rgStores[ii+1].cItems));
			AssertSameSize(pAppInfo->rgbitStores[ii], g_rgStores[ii+1].bit);
		DmWrite(pAppInfo, OffsetOf(*pAppInfo, rgbitStores[ii]), &g_rgStores[ii+1].bit, sizeof(g_rgStores[ii+1].bit));
		if (g_rgStores[ii+1].bit)
			{
			Assert(StrLen(g_rgStores[ii+1].pszName) <= cchMaxStoreName);
			DmStrCopy(pAppInfo, OffsetOf(*pAppInfo, rgszStoreNames[ii]), g_rgStores[ii+1].pszName);
			}
		}
	MemPtrUnlock(pAppInfo);

	// Write the preferences / saved-state information.
	MemSet(pprefs, sizeof(*pprefs), 0);
	pprefs->irecTopVisible = irecTopVisible;
	pprefs->catCurrent = catCurrent;
	pprefs->view = g_view;
	pprefs->store = g_store;
	pprefs->columns = g_columns;

	pprefs->flags = 0;
	if (g_fOnceDefault)
		pprefs->flags |= prefOneTime;
	if (g_fDecimals)
		pprefs->flags |= prefUseDecimals;
	if (g_fAutoQuantity)
		pprefs->flags |= prefAutoQuantity;

	// Write the state information.
	PrefSetAppPreferences(AppType, 1, PrefsVersionNum, pprefs,
						  sizeof(*pprefs), fTrue);

Error:
	// Close the application's database.
	DmCloseDatabase(g_pdb);

	if (pprefs)
		MemPtrFree(pprefs);
	if (g_pSysPrefs)
		MemPtrFree(g_pSysPrefs);

	for (ii = g_cStores; ii--;)
		if (g_rgStores[ii+1].pszName)
			MemPtrFree(g_rgStores[ii+1].pszName);

	__ListLeaks();
}


static void Search(FindParamsPtr findParams)
{
	DmOpenRef pdb;
	LocalID dbID;
	UInt cardNo = 0;
	VoidHand hHeader;
	CharPtr pszHeader;
	Boolean fHidePrivate;
	Boolean fDone;
	UInt record;
	VoidHand hRecord;
	Item *pitem;
	Boolean fMatch;
	Word pos;
	RectangleType rc;
	UInt cch;
	UInt cchMin;
	Char sz[128];

	fHidePrivate = PrefGetPreference(prefHidePrivateRecords);

	// Find the application's data file.
	pdb = DmOpenDatabaseByTypeCreator(DBType, AppType, findParams->dbAccesMode);
	if (!pdb)
		{
		// Return without anything.  Also indicate that no more records are expected.
		findParams->more = fFalse;
		goto Error;
		}
	DmOpenDatabaseInfo(pdb, &dbID, 0, 0, &cardNo, 0);

	// Display the heading line.
	hHeader = DmGetResource(strRsc, STR_FINDHEADER);
	pszHeader = MemHandleLock(hHeader);
	fDone = FindDrawHeader(findParams, pszHeader);
	MemHandleUnlock(hHeader);
	if (fDone)
		goto Error;			// no more room to display the heading line

	// Search the records for the string.  We start with the recordNum passed.
	// This allows the search code to be called multiple times when more than
	// one screen of records are found.
	for (record = findParams->recordNum; !fDone; ++record)
		{
		// get next record (skip private records if necessary)
		hRecord = DmQueryNextInCategory(pdb, &record, dmAllCategories);

		// Have we run out of records?
		if (!hRecord)
			{
			findParams->more = fFalse;
			break;
			}

		pitem = MemHandleLock(hRecord);
		if (fHidePrivate && (pitem->flags & flagPrivate))
			goto EndLoop;

		// Search for the string passed,  if it's found display the title
		// of the memo.  An application should call FindStrInStr
		// for each part of the record which needs to be searched.
		fMatch = FindStrInStr(pitem->szName, findParams->strToFind, &pos);
		if (fMatch)
			{
			// Add the match to the find parameter block, if there is no room
			// to display the match the following function will return true.
			fDone = FindSaveMatch(findParams, record, pos, 0, 0, cardNo, dbID);

			if (!fDone)
				{
				// Get the bounds of the region where we will draw the results.
				FindGetLineBounds(findParams, &rc);

				// Display the record in the search dialog.  We move in one pixel
				// so that when the record is inverted the left edge is solid.
				rc.topLeft.x++;
				rc.extent.x--;
				//$ review: (chrisant) this is a lot of work just for an
				// ellipsis.  could just as easily get rid of sz and directly
				// call FntCharsInWidth to find number of chars WinDrawChars
				// should draw.
				cch = StrLen(pitem->szName);
				cchMin = min(cch, DimensionOf(sz)-1);
				StrNCopy(sz, pitem->szName, cchMin);
				sz[cchMin] = 0;
				if (cch > cchMin)
					sz[cchMin-1] = horizEllipsisChr;
				cch = TruncateLabel(sz, rc.extent.x);
				WinDrawChars(sz, cch, rc.topLeft.x, rc.topLeft.y);

				// increment find view line number
				findParams->lineNumber++;
				}
			}
EndLoop:
		MemHandleUnlock(hRecord);
		}

Error:
	if (pdb)
		DmCloseDatabase(pdb);
}


static void GoToRecord(GoToParamsPtr pGoToParams, Boolean fLaunchingApp)
{
	Word		record;
	UInt		attr;
	ULong		uniqueID;
	EventType	event;
	VoidHand	h;
	Item		*pitem;

	record = pGoToParams->recordNum;
	DmRecordInfo(g_pdb, record, &attr, &uniqueID, NULL);

	// change current category if necessary
	if (catCurrent != dmAllCategories)
		catCurrent = attr & dmRecAttrCategoryMask;

	// If the application is already running, close all open forms.
	// Because closing forms sometimes writes data to a record which
	// could affect the ordering of records in the database, the record
	// is looked up by it's unique id which is unaffected by changing
	// data and database ordering.
	if (!fLaunchingApp)
		{
		FrmCloseAllForms();
		DmFindRecordByID(g_pdb, uniqueID, &record);
		}

	h = DmQueryRecord(g_pdb, record);
	if (h)
		{
		pitem = MemHandleLock(h);

		// change current store if necessary
		if (g_store && !(pitem->store & g_store))
			g_store = pitem->store & ~storeOnce;
		// change current view if necessary
		if (g_view == viewCoupon && !(pitem->flags & flagCoupon))
			g_view = viewNeed;
		if (g_view == viewNeed && !(pitem->flags & flagNeed))
			g_view = viewAll;

		MemHandleUnlock(h);
		}
	else
		Error(errErrorOpeningRecord);

	// Send events to load and goto a form selecting the matching text.
	MemSet(&event, sizeof(EventType), 0);

	event.eType = frmLoadEvent;
	event.data.frmLoad.formID = FRM_MAIN;
	EvtAddEventToQueue(&event);

	if (h)
		{
		event.eType = frmGotoEvent;
		event.data.frmGoto.recordNum = record;
		event.data.frmGoto.matchPos = pGoToParams->matchPos;
		event.data.frmGoto.matchLen = pGoToParams->searchStrLen;
		event.data.frmGoto.matchFieldNum = pGoToParams->matchFieldNum;
		event.data.frmGoto.formID = FRM_MAIN;
		EvtAddEventToQueue(&event);
		}
}


static Err ECreateRecord(void)
{
	Err			err = 0;
	Ptr			p;
	VoidHand	hRecord;
	UInt		i = firstRecord;
	Char		szNull[1] = {0};
	UInt		attr;			// P11. record attributes (category)
	Item		item;
	UChar		aisle = 1;
	ItemPtr		pitem;

	if (s_record != noRecordSelected)
		{
		hRecord = DmQueryRecord(g_pdb, s_record);
		if (hRecord)
			{
			pitem = MemHandleLock(hRecord);
			if (pitem)
				{
				aisle = pitem->aisle;
				MemHandleUnlock(hRecord);
				}
			}
		}

	// Create a new first record in the database
	hRecord = DmNewRecord(g_pdb, &i, sizeof(Item) + sizeof(szNull));
	if (!hRecord)
		{
		err = DmGetLastErr();
		Assert(err);
		goto Error;
		}

	// Lock down the block containing the new record.
	p = MemHandleLock(hRecord);

	// initialize new item
	MemSet(&item, sizeof(item), 0);
	if (g_view != viewAll)
		{
		item.uQuantity = 1;
		if (g_view == viewCoupon)
			item.flags |= flagCoupon;

		item.flags |= flagNeed;
		UpdateCounts(1, g_store);
		}
	item.aisle = aisle;
	item.store = g_store;
	if (g_fOnceDefault)
		item.store |= storeOnce;
	CE(DmWrite(p, 0, &item, sizeof(item)));
	CE(DmStrCopy(p, sizeof(item), szNull));

	// Unlock the block of the new record.
	MemPtrUnlock(p);

	// WARNING: (chrisant) be careful about using CE() after this point, watch
	// out for DmReleaseRecord below.
	CE(DmRecordInfo(g_pdb, i, &attr, NULL, NULL));
	attr &= ~dmRecAttrCategoryMask;				// Remove all category bits
	if (catCurrent == dmAllCategories)
		{
		// Since the user isn't looking in any particular category place
		// the new record in the unfiled category.
		attr |= dmUnfiledCategory;
		}
	else
		{
		// Set the attribute bits to indicate the current category.
		attr |= catCurrent;
		}
	//$ review: (chrisant) no error handling here.  watch out for
	// DmReleaseRecord below if you add CE() here.
	DmSetRecordInfo(g_pdb, i, &attr, NULL);

	// Release the record to the database manager.  The true value indicates that
	// the record contains "dirty" data.  Release Record will set the record's
	// dirty flag and update the database modification count.
	DmReleaseRecord(g_pdb, i, fTrue);	// dmErrIndexOutOfRange is only error

	// Remember the index of the current record.
	SetCurrentRecord(i);

	// set graffiti state for an initial upshift of first character
	GrfSetState(fFalse, fFalse, fTrue);

Error:
	return err;
}



///////////////////////////////////////////////////////////////////////////
// Table Routines

static void MainViewUpdateScrollers(FormPtr pfrm, Int recBottom)
{
	Word			record;
	Boolean			fScrollableUp;
	Boolean			fScrollableDown;
	TablePtr		ptbl;
	Word			row;
	RectangleType	rcItem;
	VoidHand		h;
	ItemPtr			pitem;

	// If the first record displayed is not the first record in the category,
	// enable the up scroller.
	record = irecTopVisible;
	fScrollableUp = F(GetFilteredRecord(g_pdb, &record, catCurrent,
										g_store, g_view, 1, dmSeekBackward));

	// If the last record displayed is not the last record in the category,
	// enable the down scroller.
	//$ review: (chrisant) this is a sucky way to ensure that the scroller is
	// enabled even if the last item is clipped.  however, it basically works,
	// and since i'm not even using a Table control in v2.0 (i'm doing it
	// myself) it's not worth fixing this the "best" way, since it would take
	// a lot of rewriting the IncrementalLoadTable etc routines.
	record = recBottom;
	ptbl = GetObjectPtr(IDC_TABLE);
	fScrollableDown = fFalse;
	if (TblFindRowData(ptbl, record, &row))
		{
		TblGetItemBounds(ptbl, row, COL_NAME, &rcItem);
		h = DmQueryRecord(g_pdb, record);
		pitem = MemHandleLock(h);
		row = CalcRowHeight(pitem->szName, rcItem.extent.x, MAX_ROWS);
		MemHandleUnlock(h);
		fScrollableDown = (Int)(row * (Word)FntLineHeight()) > rcItem.extent.y;
		}
	if (!fScrollableDown)
		fScrollableDown = F(GetFilteredRecord(g_pdb, &record, catCurrent,
											  g_store, g_view, 1, dmSeekForward));

	// Update the scroll button.
	FrmUpdateScrollers(pfrm, FrmGetObjectIndex(pfrm, IDC_UP),
					   FrmGetObjectIndex(pfrm, IDC_DOWN),
					   fScrollableUp, fScrollableDown);
}


void TableInitPrice(TablePtr ptbl, FieldPtr pfld, Word row, Word column,
					RectanglePtr prc, Boolean fEditing)
{
	VoidHand	hText;
	UInt		record;
	VoidHand	hRecord;
	ItemPtr		pitem;
	
	MemSet(pfld, sizeof(*pfld), 0);
	
	RctCopyRectangle(prc, &pfld->rect);

	// zzz use API routines
	pfld->attr.usable = fTrue;
	pfld->attr.visible = fEditing;
	pfld->attr.editable = fTrue;
	pfld->attr.singleLine = fTrue;
	pfld->attr.dynamicSize = fFalse;
	pfld->attr.underlined = fTrue;
	pfld->attr.insPtVisible = fTrue;
	pfld->attr.numeric = fTrue;
	pfld->attr.justification = rightAlign;
	
	pfld->maxChars = cchMaxPriceText;

	record = TblGetRowData(ptbl, row);
	hRecord = DmQueryRecord(g_pdb, record);
	Assert(hRecord);
	pitem = MemHandleLock(hRecord);
	Assert(pitem);

	if (pitem && pitem->ulPrice)
		{
		hText = MemHandleNew(cchMaxPriceText + 1);
		StrCopy(MemHandleLock(hText),
				LocalizeNumber(pitem->ulPrice, fFalse, fEditing, 0));
		MemHandleUnlock(hText);
		FldSetTextHandle(pfld, (Handle)hText);
		FldDrawField(pfld);
		}

	MemHandleUnlock(hRecord);
}


static void TableSelectPrice(EventPtr pEvent, TablePtr ptbl, Word row)
{
	FieldPtr pfld;
	EventType newEvent;
	RectangleType rc;

	g_fEditingPrice = fTrue;
	TblGetItemBounds(ptbl, row, COL_PRICE, &rc);

	// If we are not already editing this item, initialize the field structure
	// that's used for editing text item and highlight the edit indication.
	if ((row != ptbl->currentRow) || 
		 (COL_PRICE != ptbl->currentColumn) ||
		 (!ptbl->attr.editing))
		{
		ptbl->attr.editing = fTrue;
		pfld = TblGetCurrentField(ptbl);

		TableInitPrice(ptbl, pfld, row, COL_PRICE, &rc, fTrue);

		ptbl->currentRow = row;
		ptbl->currentColumn = COL_PRICE;

		InvertEditIndicator(ptbl, row);
		}

	// convert the table enter event to a field enter event.
	pfld = TblGetCurrentField(ptbl);
	EvtCopyEvent(pEvent, &newEvent);
	newEvent.eType = fldEnterEvent;
	newEvent.data.fldEnter.fieldID = pfld->id;
	newEvent.data.fldEnter.pField = pfld;
#if 0
	if (!g_fDecimals)
		{
		FontID font;

		font = FntSetFont(pfld->fontID);
		newEvent.screenX -= FntCharsWidth(".00");
		FntSetFont(font);
		}
#endif

	FldHandleEvent(pfld, &newEvent);
}


void TableSavePrice(TablePtr ptbl)
{
	FieldPtr		pfld;
	UInt			record;
	VoidHand		hRecord;
	CharPtr			psz;
	ULong			ulPrice = 0;
	ULong			ul = 0;
	Int				cDecimal = 0;
	Item			*pitem;
	Word			row;
	RectangleType	rc;

	Assert(g_pdb);
	Assert(ptbl);
	AssertSameSize(pitem->ulPrice, ulPrice);

	if (!g_fEditingPrice)
		return;
	g_fEditingPrice = fFalse;

	Assert(recCurrent != noRecordSelected);
	TblFindRowData(ptbl, recCurrent, &row);

	// if the field is not dirty, we've nothing to do
	pfld = TblGetCurrentField(ptbl);
	Assert(pfld);
	if (!FldDirty(pfld))
		goto Out;

	// Get the record number that corresponds to the table item to load.
	record = TblGetRowData(ptbl, row);
	Assert(record == recCurrent);	// i think this is true, but am not confident enough to assume that

	hRecord = DmGetRecord(g_pdb, record);

	// parse price
	//$ review: (chrisant) yikes seems pretty inefficient
	for (psz = FldGetTextPtr(pfld); psz && *psz; ++psz)
		{
		if ((*psz) >= '0' && (*psz) <= '9')
			{
			ul *= 10;
			ul += ((*psz) - '0');
			if (cDecimal)
				{
				cDecimal++;
				if (cDecimal == 3)
					{
					// round based on third digit after decimal
					//$ review: (chrisant) do cash registers always round up
					// from 5?  should we use statistical rounding instead?
					if (psz[1] >= '5' && psz[1] <= '9')
						ul++;
					break;
					}
				}
			}
		else if (*psz == g_chDecimal && !cDecimal)
			{
			ulPrice = ul;
			ul = 0;
			cDecimal = 1;
			}
		else if (!(*psz == g_chSep || *psz == '$'))
			break;
		}
	if (!cDecimal)
		{
		ulPrice = ul;
		ul = 0;
		}
	ulPrice *= 100;
	if (cDecimal == 2)
		ulPrice += ul * 10;
	else if (cDecimal == 3)
		ulPrice += ul;

	// set new price
	pitem = MemHandleLock(hRecord);
	DmWrite(pitem, OffsetOf(*pitem, ulPrice), &ulPrice, sizeof(pitem->ulPrice));
	MemHandleUnlock(hRecord);

	DmReleaseRecord(g_pdb, record, fTrue);

//Error:
Out:
	FldReleaseFocus(pfld);
	FldSetSelection(pfld, 0, 0);
	InvertEditIndicator(ptbl, row);
	ptbl->attr.editing = fFalse;

	TblGetItemBounds(ptbl, row, COL_PRICE, &rc);
	WinEraseRectangle(&rc, 0);
	TblCallback_DrawPrice(ptbl, row, COL_PRICE, &rc);
}


static Err MainViewLoadDataCallback(VoidPtr ptbl, Word row, Word column,
									Boolean editable, VoidHand *phData,
									WordPtr pdataOffset, WordPtr pdataSize)
{
	Err				err = 0;
	VoidHand		hRecord;
	UInt			record;
	Item			*pitem;
	FieldPtr		pfld;

	CALLBACK_PROLOGUE

	Assert(g_pdb);
	Assert(ptbl);

	// init field
	pfld = TblGetCurrentField(ptbl);
#ifdef DEBUG
	if (editable)
		Assert(pfld);
#endif

	// Get the record number that corresponds to the table item to load.
	record = TblGetRowData(ptbl, row);

	// Retrieve the record from the database and lock it down.
	hRecord = DmQueryRecord(g_pdb, record);
	Assert(hRecord);
	if (!hRecord)
		CE(DmGetLastErr());

	pitem = MemHandleLock(hRecord);
	Assert(pitem);

	*pdataOffset = OffsetOf(*pitem, szName);
	//*pdataSize = pitem->cchName + 1;
	*pdataSize = StrLen(pitem->szName) + 1;
	*phData = hRecord;

//$ review: (chrisant) the assert was actually firing when the field resized to multiple lines!
#if 0 //def DEBUG
	Assert(pitem->cchName == StrLen(pitem->szName));
	if (pitem->cchName != StrLen(pitem->szName))
		*pdataSize = StrLen(pitem->szName) + 1;
#endif

	// Unlock the handle to the record.
	MemHandleUnlock(hRecord);

	if (editable && pfld)
		{
		FieldAttrType attr;

		FldGetAttributes(pfld, &attr);
		attr.autoShift = fTrue;
		attr.singleLine = fFalse;
		attr.dynamicSize = fTrue;
		attr.numeric = fFalse;
		FldSetAttributes(pfld, &attr);
		}

Error:
	CALLBACK_EPILOGUE
	return err;
}


static Boolean MainViewSaveDataCallback(VoidPtr ptbl, Word row, Word column)
{
	FieldPtr	pfld;
	UInt		record;
	VoidHand	hRecord;
	Item		*pitem;
	UChar		cchName;
	Boolean		fRedraw = fFalse;

	CALLBACK_PROLOGUE

	Assert(g_pdb);
	Assert(ptbl);
	Assert(sizeof(cchName) == sizeof(pitem->cchName));

	pfld = TblGetCurrentField(ptbl);
	record = TblGetRowData(ptbl, row);

	// if the field is not dirty, we've nothing to do
	if (!FldDirty(pfld))
		goto Out;

	hRecord = DmGetRecord(g_pdb, record);
	Assert(hRecord);
	if (hRecord)
		{
		pitem = MemHandleLock(hRecord);
		cchName = FldGetTextLength(pfld);
		DmWrite(pitem, OffsetOf(*pitem, cchName), &cchName, sizeof(pitem->cchName));
		MemHandleUnlock(hRecord);

		DmReleaseRecord(g_pdb, record, fTrue);
		fRedraw = fTrue;
		}

Out:
	CALLBACK_EPILOGUE
	TblMarkRowInvalid(ptbl, row);
	return fRedraw;
}


static void TblCallback_DrawQuantity(VoidPtr ptbl, Word row, Word column, RectanglePtr prc)
{
	Word len;
	short width;
	Char sz[8];

	CALLBACK_PROLOGUE

	FntSetFont(boldFont);

	StrIToA(sz, TblGetItemInt(ptbl, row, column));
	len = StrLen(sz);
	if (len > 2)
		{
		StrCopy(sz, "+");
		len = 1;
		}

	width = FntCharsWidth(sz, len) - 1;
	WinDrawChars(sz, len,
				 prc->topLeft.x + ((((TablePtr)ptbl)->columnAttrs[column].width - width) / 2),
				 prc->topLeft.y);

	CALLBACK_EPILOGUE
}


static void TblCallback_DrawCategory(VoidPtr ptbl, Word row, Word column, RectanglePtr prc)
{
	Char	sz[dmCategoryLength];
	UInt	record;
	UInt	cat;
	UInt	width;
	UInt	len;
	Boolean	fFit;

	CALLBACK_PROLOGUE

	//$ review: (chrisant) this is probably going to be pretty slow...
	//$ todo: (chrisant) store category as ItemInt?
	record = TblGetRowData(ptbl, row);
	CEv(DmRecordInfo(g_pdb, record, &cat, NULL, NULL));
	cat &= dmRecAttrCategoryMask;
	CategoryGetName(g_pdb, cat, sz);

	FntSetFont(stdFont);
	//width = TblGetColumnWidth(ptbl, column);
	width = prc->extent.x;
	len = StrLen(sz);
	FntCharsInWidth(sz, &width, &len, &fFit);
	WinDrawChars(sz, len, prc->topLeft.x, prc->topLeft.y);

Error:
	;
	CALLBACK_EPILOGUE
}


static void TblCallback_DrawPrice(VoidPtr ptbl, Word row, Word column, RectanglePtr prc)
{
	Word y;
	FontID curFont;
	FieldType field;

	CALLBACK_PROLOGUE
	
	TableInitPrice(ptbl, &field, row, column, prc, fFalse);

	FldDrawField(&field);
	FldFreeMemory(&field);

	// zzz There a bug in the API that prevents right-aligned fields 
	// from drawing correctly.
	curFont = FntSetFont(FldGetFont(&field));
	y = prc->topLeft.y + FntCharHeight() - 1;
	WinDrawGrayLine(prc->topLeft.x, y, prc->topLeft.x + prc->extent.x - 1, y);
	FntSetFont(curFont);

	CALLBACK_EPILOGUE
}


static Boolean SetNeed(Item *pitem, Boolean fNeed)
{
	UInt uQuantity;
	Boolean fChanged;
	UChar flags;

	Assert(sizeof(pitem->flags) == sizeof(flags));
	Assert(sizeof(pitem->uQuantity) == sizeof(uQuantity));

	flags = (pitem->flags & ~flagNeed);
	if (fNeed)
		flags |= flagNeed;
	fChanged = (flags != pitem->flags);
	if (fChanged)
		{
		uQuantity = pitem->uQuantity;
		UpdateCounts(fNeed ? 1 : -1, pitem->store);
		if (g_fAutoQuantity)
			uQuantity = fNeed ? 1 : 0;
		DmWrite(pitem, OffsetOf(*pitem, flags), &flags, sizeof(pitem->flags));
		DmWrite(pitem, OffsetOf(*pitem, uQuantity), &uQuantity, sizeof(pitem->uQuantity));
		}
	return fChanged;
}


static Boolean ToggleItemFlag(TablePtr ptbl, Int row, Int column)
{
	Boolean			f = fFalse;
	Int				record;
	VoidHand		hRecord;
	Item			*pitem;
	UChar			flags;

	Assert(sizeof(pitem->flags) == sizeof(flags));

	record = TblGetRowData(ptbl, row);
	hRecord = DmGetRecord(g_pdb, record);
	pitem = MemHandleLock(hRecord);

	switch (column)
		{
		case COL_NEED:
			f = !(pitem->flags & flagNeed);
			SetNeed(pitem, f);
			TblSetItemInt(ptbl, row, COL_QUANTITY, pitem->uQuantity);
			TblMarkRowInvalid(ptbl, row);
			break;
		case COL_COUPON:
			flags = pitem->flags ^ flagCoupon;
			f = F(flags & flagCoupon);
			DmWrite(pitem, OffsetOf(*pitem, flags), &flags, sizeof(pitem->flags));
			break;
		default:
			Panic();
		}

	MemHandleUnlock(hRecord);
	DmReleaseRecord(g_pdb, record, fTrue);

	TblSetItemInt(ptbl, row, column, f);

	TblRedrawTable(ptbl);
	return f;
}


static void LoadTableRow(TablePtr ptbl, Word row, VoidHand h, UInt record)
{
	Item			*pitem;
	Boolean			f;
	RectangleType	rc;
	Int				cy;
	Int				cyLine = FntLineHeight();

	pitem = MemHandleLock(h);
	Assert(pitem);

	TblSetRowData(ptbl, row, record);

	// Need/Purchased
	TblSetItemStyle(ptbl, row, COL_NEED, checkboxTableItem);
	f = (pitem->flags & flagNeed) ? fTrue : fFalse;
	TblSetItemInt(ptbl, row, COL_NEED, (g_view == viewAll) ? f : !f);

	// Description
	TblSetItemStyle(ptbl, row, COL_NAME, textTableItem);
	TblGetItemBounds(ptbl, row, COL_NAME, &rc);
	#ifdef DEBUG
	cy = TblGetRowHeight(ptbl, row);
	Assert(cy == rc.extent.y);
	#endif

	cy = CalcRowHeight(pitem->szName, rc.extent.x, g_cLinesRemaining);
	g_cLinesRemaining -= cy;
	g_cVisibleItems++;
	TblSetRowHeight(ptbl, row, cy * cyLine);

	// Quantity
	//TblSetItemStyle(ptbl, row, COL_QUANTITY, numericTableItem);
	TblSetItemStyle(ptbl, row, COL_QUANTITY, customTableItem);
	TblSetItemInt(ptbl, row, COL_QUANTITY, pitem->uQuantity);

	// Coupon
	TblSetItemStyle(ptbl, row, COL_COUPON, checkboxTableItem);
	TblSetItemInt(ptbl, row, COL_COUPON, pitem->flags & flagCoupon);

	// Aisle
	TblSetItemStyle(ptbl, row, COL_AISLE, numericTableItem);
	TblSetItemInt(ptbl, row, COL_AISLE, pitem->aisle);

	// Category
	TblSetItemStyle(ptbl, row, COL_CATEGORY, customTableItem);
//	TblSetItemInt(ptbl, row, COL_CATEGORY, pitem->);

	// Price
	TblSetItemStyle(ptbl, row, COL_PRICE, customTableItem);

	TblSetRowUsable(ptbl, row, fTrue);
	TblSetRowSelectable(ptbl, row, fTrue);

	MemHandleUnlock(h);
}


static void MainViewLoadTable(FormPtr pfrm, TablePtr ptbl, UInt record)
{
	Word			row;
	Word			recordLast = firstRecord;
	Word			cRows;
	VoidHand		hNext;
	FontID			font;

	Assert(pfrm);
	Assert(ptbl);

TryAgain:
	font = FntSetFont(stdFont);

	// For each row in the table, store the record number in the
	// table item that will display the record.
	g_cVisibleItems = 0;
	g_cLinesRemaining = cRows = TblGetNumberOfRows(ptbl);

	for (row = 0; row < cRows; row++)
		{
		hNext = 0;
		if (g_cLinesRemaining)
			{
			// Get the next record in the current category.
			hNext = GetFilteredRecord(g_pdb, &record, catCurrent,
									  g_store, g_view,
									  (row == 0) ? 0 : 1, dmSeekForward);
			}

		// If the record was found, store the record number in the table item,
		// otherwise set the table row unusable.
		if (hNext)
			{
			LoadTableRow(ptbl, row, hNext, record);
			recordLast = record;
			}
		else
			{
			// Sometimes there are more table rows than records.
			// If this happens mark those unused table rows as
			// not usable.
			TblSetRowUsable(ptbl, row, fFalse);
			}

		// Mark the row invalid so that it will draw when we call the
		// draw routine.
		TblMarkRowInvalid(ptbl, row);
		}

	FntSetFont(font);

	if (recCurrent != noRecordSelected &&
			(recCurrent < irecTopVisible || recCurrent > recordLast))
		{
		// since one or more items may take up more than a single line, this
		// makes sure recCurrent doesn't fall off the bottom visible edge of
		// the table.
		if (recCurrent == irecTopVisible)
			{
			// this avoids infinite loop
			//$ todo: (chrisant) use SetCurrentRecord?
			recCurrent = noRecordSelected;
			irecTopVisible = firstRecord;
			}
		else
			{
			irecTopVisible = recCurrent;
			}
		record = irecTopVisible;

		PanicSz("trying again...");
		goto TryAgain;
		}

	// P13. Update the scroll arrows.
	MainViewUpdateScrollers(pfrm, recordLast);
}


static void MainViewLoadRecords(FormPtr pfrm)
{
	TablePtr	ptbl;

	Assert(pfrm);
	Assert(g_pdb);

	// Get the table and it's number of rows
	ptbl = FrmGetObjectPtr(pfrm, FrmGetObjectIndex(pfrm, IDC_TABLE));
	Assert(ptbl);

	// Before we load the table with records we should do any positioning
	// of the table now.

	// If there is a current record is it visible?
	if (recCurrent != noRecordSelected)
		{
		// Is the current record before the first visible record?
		if (irecTopVisible > recCurrent)
			{
			// Make recCurrent the top record displayed.
			irecTopVisible = recCurrent;
			}
		else
			{
			// Is the current record after the last visible record?
			if (!g_cVisibleItems ||
				recCurrent > TblGetRowData(ptbl, g_cVisibleItems - 1))
				{
				irecTopVisible = recCurrent;
				}
			}
		}

	// Now that the table is in position load the records.
	MainViewLoadTable(pfrm, ptbl, irecTopVisible);
}


static void ReloadTable(TablePtr ptbl)
{
	TblReleaseFocus(ptbl);
	MainViewLoadRecords(FrmGetActiveForm());
	TblEraseTable(ptbl);
	TblDrawTable(ptbl);
}


static void IncrementalLoadTable(TablePtr ptbl)
{
	Word			cyLine = FntLineHeight();
	Word			cy;
	Word			row = 0;
	UInt			record;
	VoidHand		h;
	Item			*pitem;
	RectangleType	rc;
	Word			cx;

	Assert(TblRowUsable(ptbl, 0));	// "incremental" is the operative word...

	TblGetItemBounds(ptbl, 0, COL_NAME, &rc);
	cx = rc.extent.x;
	g_cVisibleItems = 0;
	g_cLinesRemaining = TblGetNumberOfRows(ptbl);
	while (fTrue)
		{
		Assert(g_cLinesRemaining >= 0);
		if (g_cLinesRemaining == 0)
			{
			// disable following rows
			for (row = TblGetNumberOfRows(ptbl); --row >= g_cVisibleItems;)
				{
				Assert(row);
				TblSetRowUsable(ptbl, row, fFalse);
				TblMarkRowInvalid(ptbl, row);
				}
			break;
			}

		if (TblRowUsable(ptbl, row))
			{
			record = TblGetRowData(ptbl, row);
			h = DmQueryRecord(g_pdb, record);
			Assert(h);
			pitem = MemHandleLock(h);
			Assert(pitem);
			cy = CalcRowHeight(pitem->szName, cx, g_cLinesRemaining);
			g_cVisibleItems++;
			g_cLinesRemaining -= cy;
			if (cy * cyLine != TblGetRowHeight(ptbl, row))
				{
				TblMarkRowInvalid(ptbl, row);
				TblSetRowHeight(ptbl, row, cy * cyLine);
				}
			MemHandleUnlock(h);
			}
		else
			{
			h = GetFilteredRecord(g_pdb, &record, catCurrent,
								  g_store, g_view, 1, dmSeekForward);
			if (!h)
				break;
			LoadTableRow(ptbl, row, h, record);
			TblMarkRowInvalid(ptbl, row);
			}

		row++;
		}
	irecTopVisible = TblGetRowData(ptbl, 0);
}


static void IncrementalLoadTableBackwards(TablePtr ptbl)
{
	Int				cyLine = FntLineHeight();
	Int				cy;
	Word			row;
	Word			cRows;
	UInt			record;
	VoidHand		h;
	Item			*pitem;
	RectangleType	rc;
	Word			cx;

	TblGetItemBounds(ptbl, 0, COL_NAME, &rc);
	cx = rc.extent.x;
	g_cVisibleItems = 0;
	cRows = g_cLinesRemaining = TblGetNumberOfRows(ptbl);
	for (row = 0; row < cRows && TblRowUsable(ptbl, row); row++)
		{
		record = TblGetRowData(ptbl, row);
		h = DmQueryRecord(g_pdb, record);
		Assert(h);
		pitem = MemHandleLock(h);
		Assert(pitem);
		cy = CalcRowHeight(pitem->szName, cx, g_cLinesRemaining);
		g_cLinesRemaining -= cy;
		g_cVisibleItems++;
		if (cy * cyLine != (Int)TblGetRowHeight(ptbl, row))
			{
			TblMarkRowInvalid(ptbl, row);
			TblSetRowHeight(ptbl, row, cy * cyLine);
			}
		MemHandleUnlock(h);
		}

	if (0 >= g_cLinesRemaining)
		return;

	for (row = 0; fTrue;)
		{
		record = TblGetRowData(ptbl, row);
		h = GetFilteredRecord(g_pdb, &record, catCurrent,
							  g_store, g_view, 1, dmSeekBackward);
		if (!h)
			break;

		pitem = MemHandleLock(h);
		Assert(pitem);
		cy = CalcRowHeight(pitem->szName, cx, MAX_ROWS);
		MemHandleUnlock(h);
		if (cy <= g_cLinesRemaining || g_cVisibleItems == 1)
			{
			Assert(row == 0);
			TblInsertRow(ptbl, row);
			LoadTableRow(ptbl, row, h, record);
			TblMarkRowInvalid(ptbl, row);
			}
		if (cy > g_cLinesRemaining)
			break;

		Assert(row == 0);	// row must stay 0
		}

	irecTopVisible = TblGetRowData(ptbl, 0);
}


static void UpdateCategoryLabel(void)
{
	Int				cx;
	RectangleType	rc;
	FormPtr			pfrm;

	pfrm = FrmGetActiveForm();
	FrmGetFormBounds(pfrm, &rc);
	cx = rc.extent.x;
	CategoryGetName(g_pdb, catCurrent, szCategoryName);
	FrmGetObjectBounds(pfrm, FrmGetObjectIndex(pfrm, IDC_VIEW_COUPON), &rc);
	CategoryTruncateName(szCategoryName, cx - (rc.topLeft.x + rc.extent.x) - 18);
	CategorySetTriggerLabel(GetObjectPtr(IDC_CATEGORY_TRIGGER), szCategoryName);
}


static Int MainViewSelectCategory(void)
{
	Word		category;
	Boolean		fEdited;
	
	// Process the category popup list.
	category = catCurrent;

	fEdited = CategorySelectV10(g_pdb, FrmGetActiveForm(), IDC_DUMMY_TRIGGER,
								IDC_CATEGORY_LIST, fTrue, &category, szCategoryName);

	if (fEdited || (category != catCurrent))
		{
		catCurrent = category;
		UpdateCategoryLabel();

		// Display the new category.
		ReloadTable(GetObjectPtr(IDC_TABLE));
		}

	return category;
}


static void MainViewScroll(DirectionType direction)
{
	TablePtr	ptbl;
	UInt		irecNewTopVisible;
	Word		row;

	Assert(!fOneLine);

	ptbl = GetObjectPtr(IDC_TABLE);
	irecNewTopVisible = irecTopVisible;
	SetCurrentRecord(noRecordSelected);

	if (direction == down)
		{
		// Scroll the table down.
#if 0
		if (fOneLine)
			{
			// Scroll down a single line.
			GetFilteredRecord(g_pdb, &irecNewTopVisible, catCurrent,
							  g_store, g_view, 1, dmSeekForward);
			}
		else
#endif
			{
			// down one page
			if (g_cVisibleItems == 1)
				GetFilteredRecord(g_pdb, &irecNewTopVisible, catCurrent,
								  g_store, g_view, 1, dmSeekForward);
			else
				{
				UInt rec;

				rec = TblGetRowData(ptbl, g_cVisibleItems - 1);
//				if (GetFilteredRecord(g_pdb, &rec, catCurrent,
//									  g_store, g_view, 1, dmSeekForward))
				if (CtlEnabled(GetObjectPtr(IDC_DOWN)))
					{
					TblReleaseFocus(ptbl);
					// clear out all rows except bottom row
					row = g_cVisibleItems - 1;
					if (row < 1)
						row = 1;
					while (row--)
						TblRemoveRow(ptbl, 0);

					// fill table
					IncrementalLoadTable(ptbl);
					IncrementalLoadTableBackwards(ptbl);
					TblEraseTable(ptbl);
					TblDrawTable(ptbl);
					MainViewUpdateScrollers(FrmGetActiveForm(),
											TblGetRowData(ptbl, g_cVisibleItems - 1));
					return;
					}
				}
			}
		}
	else
		{
		// Scroll the table up.
#if 0
		if (fOneLine)
			{
			// Scroll up a single line
			GetFilteredRecord(g_pdb, &irecNewTopVisible, catCurrent,
							  g_store, g_view, 1, dmSeekBackward);
			}
		else
#endif
			{
			// up one page
			irecNewTopVisible = firstRecord;
			GetFilteredRecord(g_pdb, &irecNewTopVisible, catCurrent,
							  g_store, g_view, 0, dmSeekForward);
			if (irecTopVisible != irecNewTopVisible)
				{
				TblReleaseFocus(ptbl);
				// clear out all rows except top row (unless only row)
				if (g_cVisibleItems > 0)
					{
					for (row = (g_cVisibleItems == 1) ? 0 : 1;
						 row < g_cVisibleItems; row++)
						TblSetRowUsable(ptbl, row, fFalse);
					}

				// fill table in reverse direction
				IncrementalLoadTableBackwards(ptbl);
				IncrementalLoadTable(ptbl);
				TblEraseTable(ptbl);
				TblDrawTable(ptbl);
				MainViewUpdateScrollers(FrmGetActiveForm(),
										TblGetRowData(ptbl, g_cVisibleItems - 1));
				return;
				}
			}
		}

	// Avoid redraw if no change
	if (irecTopVisible != irecNewTopVisible)
		{
		// The table should be at a different position.  Load the
		// table with new records and redraw it.
		irecTopVisible = irecNewTopVisible;
		ReloadTable(ptbl);
		}
}


static const struct
	{
	Char col;
	Char spacing;
	Char width;
	Char ofs;
	} c_colinfo[MAX_COLUMNS] = { {colNeed, 2, 12, 0},
								 {colName, 2, 0, 0},
								 {colQuantity, 2, 12, 3},
								 {colCoupon, 2, 12, 3},
								 {colAisle, 2, 12, 4},
								 {colCategory, 2, 20, 1},
								 {colPrice, 1, 28, 1},
								 };

static void MainViewInit(void)
{
	FormPtr pfrm;
	TablePtr ptbl = GetObjectPtr(IDC_TABLE);
	ControlPtr pctl;
	Boolean f;
	Int ii;
	Int cx;
	Int cxGrowable;
	FontID font;
	RectangleType rc;
	CharPtr psz;
	Int rgx[MAX_COLUMNS];

	DiscardSavedInsPt();

	// Get a pointer to the main form.
	pfrm = FrmGetActiveForm();
	Assert(pfrm);
	FrmGetFormBounds(pfrm, &rc);
	cxGrowable = rc.extent.x;

	CtlSetValue(GetObjectPtr(IDC_VIEW_ALL + g_view), fTrue);

	// set category trigger label
	UpdateCategoryLabel();

	// initialize columns
	TblSetLoadDataProcedure(ptbl, COL_NAME, MainViewLoadDataCallback);
	TblSetSaveDataProcedure(ptbl, COL_NAME, MainViewSaveDataCallback);
	// custom draw procedure for Quantity (when >99, show "+")
	TblSetCustomDrawProcedure(ptbl, COL_QUANTITY, TblCallback_DrawQuantity);
	TblSetCustomDrawProcedure(ptbl, COL_CATEGORY, TblCallback_DrawCategory);
	TblSetCustomDrawProcedure(ptbl, COL_PRICE, TblCallback_DrawPrice);
	Assert(g_columns);
	for (ii = MAX_COLUMNS; ii--;)
		{
		Assert(c_colinfo[ii].col == 1 << ii);

		f = F(g_columns & c_colinfo[ii].col);

		cx = f ? c_colinfo[ii].spacing : 0;
		TblSetColumnSpacing(ptbl, ii, cx);
		cxGrowable -= cx;

		cx = f ? c_colinfo[ii].width : 0;
		TblSetColumnWidth(ptbl, ii, cx);
		cxGrowable -= cx;

		TblSetColumnUsable(ptbl, ii, f);

		rgx[ii] = cxGrowable + c_colinfo[ii].ofs;
		if (colPrice == c_colinfo[ii].col)
			{
			pctl = GetObjectPtr(IDC_PRICE);
			//$ review: (chrisant) whew, a bit little dangerous, but it works on PalmOS 2.x and 3.x
			pctl->bounds.topLeft.x = rgx[ii];
			}
		}
	Assert(cxGrowable);
	TblSetColumnWidth(ptbl, COL_NAME, cxGrowable);

	// stores trigger
	g_cxMaxStorePixels = cxGrowable - 3;
	SetStoreTrigger();

	// load data
	MainViewLoadRecords(pfrm);

	// draw form
	FrmDrawForm(pfrm);
	if (g_columns & colPrice)
		CtlShowControl(GetObjectPtr(IDC_PRICE));

	// draw fancy (i) help button in title bar
	DrawHelpButton(pfrm, IDC_HELP);

	// draw dynamic column headers
	for (ii = MAX_COLUMNS; ii--;)
		{
		if (F(g_columns & c_colinfo[ii].col))
			{
			psz = 0;

			switch (c_colinfo[ii].col)
				{
				case colQuantity:
					IfGerman(psz = "Q";)
					IfEnglish(psz = "Q";)
					break;
				case colCoupon:
					IfGerman(psz = "G";)
					IfEnglish(psz = "C";)
					break;
				case colAisle:
					IfGerman(psz = "G";)
					IfEnglish(psz = "A";)
					break;
				case colCategory:
					IfGerman(psz = "Kat";)
					IfEnglish(psz = "Cat";)
					break;
				}
			if (psz)
				{
				font = FntSetFont(stdFont);
				WinDrawChars(psz, StrLen(psz), rgx[ii], 16);
				FntSetFont(font);
				}
			}
		}
}



///////////////////////////////////////////////////////////////////////////
// MainFormHandleEvent

static void DoQuantity(TablePtr ptbl, UInt record, Int row, Int col)
{
	ListPtr			plst;
	Int				iList;
	VoidHand		h;
	Item			*pitem;
	UInt			uQuantity;
	RectangleType	rc;

	Assert(sizeof(uQuantity) == sizeof(pitem->uQuantity));

	plst = GetObjectPtr(IDC_QUANTITY_LIST);

	TblUnhighlightSelection(ptbl);

	LstSetSelection(plst, -1);		//$ review: (chrisant) set to actual?
	TblGetItemBounds(ptbl, row, col, &rc);
	LstSetPosition(plst, rc.topLeft.x, rc.topLeft.y);
	LstSetHeight(plst, 12);
	iList = LstPopupList(plst);
	if (iList == 11)
		{
		Assert(recCurrent == record);
		FrmPopupForm(FRM_QUANTITY);
		}
	else if (iList != -1)
		{
		h = DmGetRecord(g_pdb, record);
		if (h)
			{
			uQuantity = iList;
			pitem = MemHandleLock(h);
			DmWrite(pitem, OffsetOf(*pitem, uQuantity), &uQuantity, sizeof(pitem->uQuantity));
			MemHandleUnlock(h);
			DmReleaseRecord(g_pdb, record, fTrue);
			TblSetItemInt(ptbl, row, col, uQuantity);
			TblMarkRowInvalid(ptbl, row);
			TblRedrawTable(ptbl);

			RestoreInsPt();
			}
		}
}


static Err DoAisle(TablePtr ptbl, UInt record, Int row, Int col)
{
	Err				err = 0;
	ListPtr			plst;
	RectangleType	rc;
	Int				iList;
	VoidHand		h;
	ItemPtr			pitem;
	UChar			aisle;

	CharPtr			*prg;
	CharPtr			psz = 0;

	Assert(sizeof(aisle) == sizeof(pitem->aisle));

	plst = GetObjectPtr(IDC_AISLE_LIST);

	prg = MemPtrNew(sizeof(CharPtr) * 100 _MEMTRACK);
	CPE(prg);
	psz = MemPtrNew(3 * 100 _MEMTRACK);
	CPE(psz);
	for (iList = 0; iList < 100;)
		{
		prg[iList] = psz;
		iList++;
		StrIToA(psz, iList);
		// point psz after null terminator
		while (*(psz++));
		}
	psz = prg[0];
	LstSetListChoices(plst, prg, 100);

	TblUnhighlightSelection(ptbl);

	LstSetSelection(plst, TblGetItemInt(ptbl, row, col)-1);
	TblGetItemBounds(ptbl, row, col, &rc);
	LstSetPosition(plst, rc.topLeft.x, rc.topLeft.y);
	LstSetHeight(plst, 100);
	iList = LstPopupList(plst);
	if (iList != -1)
		{
		h = DmGetRecord(g_pdb, record);
		if (h)
			{
			aisle = (UChar)iList + 1;
			pitem = MemHandleLock(h);
			DmWrite(pitem, OffsetOf(*pitem, aisle), &aisle, sizeof(pitem->aisle));
			MemHandleUnlock(h);
			DmReleaseRecord(g_pdb, record, fTrue);
			TblSetItemInt(ptbl, row, col, aisle);
			TblMarkRowInvalid(ptbl, row);
			TblRedrawTable(ptbl);

			RestoreInsPt();
			}
		}

Error:
	if (prg)
		MemPtrFree(prg);
	if (psz)
		MemPtrFree(psz);
	return err;
}


static void DrawStoreName(Word iItem, RectanglePtr prc, CharPtr *rgpsz)
{
	FontID		font;
	UInt		cItems;
	Char		sz[cchMaxStoreName + 1];
	Int			cch;

	CALLBACK_PROLOGUE

	if (iItem <= g_cStores)
		cItems = g_rgStores[iItem].cItems;
	else
		cItems = 0;

	Assert(StrLen(rgpsz[iItem]) <= cchMaxStoreName);

	font = FntSetFont((cItems > 0) ? boldFont : stdFont);
	StrCopy(sz, rgpsz[iItem]);
	cch = TruncateLabel(sz, prc->extent.x);
	WinDrawChars(sz, cch, prc->topLeft.x, prc->topLeft.y);
	FntSetFont(font);

	CALLBACK_EPILOGUE
}


static void Lookup(void)
{
//	static Char		szLookup[cchMaxLookup+1];
	TablePtr		ptbl = GetObjectPtr(IDC_TABLE);
	FieldPtr		pfld = GetObjectPtr(IDC_LOOKUP);
	Char			sz[cchMaxLookup+1];
	Word			cch;
	UInt			record;
	VoidHand		h;
	Item			*pitem;
	Word			row;
	Int				compare;
	Boolean			fFound = fFalse;
	Int				offset = 0;

	Assert(pfld);
	Assert(!StrLen(_FldGetTextPtr(pfld)) < sizeof(sz));

	StrCopy(sz, _FldGetTextPtr(pfld));
	cch = StrLen(sz);

	// no-op if no text to search for
	if (*sz)
		{
		for (record = firstRecord; fTrue;)
			{
			h = GetFilteredRecord(g_pdb, &record, catCurrent,
								  g_store, g_view, offset, dmSeekForward);
			if (!h)
				break;
			offset = 1;
			
			pitem = MemHandleLock(h);
			compare = StrNCaselessCompare(sz, pitem->szName, cch);
			MemHandleUnlock(h);

			if (compare == 0)
				{
				// found match
				if (!TblFindRowData(ptbl, record, &row))
					{
					// not in table, must reload table
					irecTopVisible = record;
					ReloadTable(ptbl);
					Verify(TblFindRowData(ptbl, record, &row));
					}
				TblSelectItem(ptbl, row, COL_NAME);
				fFound = fTrue;
				break;
				}
			}
		}
	if (!fFound)
		TblUnhighlightSelection(ptbl);
}


static Int CompareItemsByName(VoidPtr pA, VoidPtr pB, Int fCompareAisles,
							  SortRecordInfoPtr psortA, SortRecordInfoPtr psortB,
							  VoidHand hAppInfo)
{
	// sorts entire database, although we could limit it to a specific store
	// and/or category fairly easily.
	if (fCompareAisles)
		{
		UChar aisleA = ((ItemPtr)pA)->aisle;
		UChar aisleB = ((ItemPtr)pB)->aisle;

		// compare aisles (if aisles are equal, fall thru and compare names)
		if (aisleA < aisleB)
			return -1;
		if (aisleA > aisleB)
			return 1;
		}
	// compare names
	return StrCompare(((ItemPtr)pA)->szName, ((ItemPtr)pB)->szName);
}


static Boolean MainViewHandleEvent(EventPtr pEvent)
{
	Int			nDir = -1;
	Boolean		fHandled = fTrue;
	TablePtr	ptbl;

	CALLBACK_PROLOGUE

	switch (pEvent->eType)
		{
		case penDownEvent:
			// set up to return focus to the record after certain actions
			SaveInsPt();

			ptbl = GetObjectPtr(IDC_TABLE);
			Assert(ptbl);
			if (recCurrent != noRecordSelected &&
				RctPtInRectangle(pEvent->screenX, pEvent->screenY,
								 &ptbl->bounds) && FCanDelete(ptbl, recCurrent))
				{
				Int ii;
				RectangleType rc;
				Word row;

				if (TblFindRowData(ptbl, recCurrent, &row))
					for (ii = MAX_COLUMNS; ii--;)
						{
						TblGetItemBounds(ptbl, row, ii, &rc);
						if (RctPtInRectangle(pEvent->screenX, pEvent->screenY,
											 &rc))
							{
							fHandled = fFalse;
							break;
							}
						}

				if (fHandled)
					{
					SetCurrentRecord(noRecordSelected);
					TblReleaseFocus(ptbl);
					DiscardSavedInsPt();
					}
				break;
				}
			fHandled = fFalse;
			break;

		case ctlExitEvent:
			// return focus to record after releasing a button w/o having
			// clicked it.
			RestoreInsPt();
			fHandled = fFalse;
			break;

		case ctlEnterEvent:
			TableSavePrice(GetObjectPtr(IDC_TABLE));
			fHandled = fFalse;
			break;

		case ctlSelectEvent:
			// a control button was pressed and released
			switch (pEvent->data.ctlEnter.controlID)
				{
				case IDC_MAIN_NEW:
					ptbl = GetObjectPtr(IDC_TABLE);

					if (FCanDelete(ptbl, recCurrent))
						DeleteRecord(recCurrent);
					recCurrent = noRecordSelected;

					// P6. Create a new database record.
					TblReleaseFocus(ptbl);
					if (FSuceeded(ECreateRecord()))
						{
						Word row;

						ReloadTable(ptbl);
						if (TblFindRowData(ptbl, recCurrent, &row))
							{
							FormPtr pfrm = FrmGetActiveForm();

							FrmSetFocus(pfrm, FrmGetObjectIndex(pfrm, ptbl->id));
							TblGrabFocus(ptbl, row, COL_NAME);
							Assert(TblGetCurrentField(ptbl));
							FldGrabFocus(TblGetCurrentField(ptbl));
							}
						#ifdef DEBUG
						else
							PanicSz("new record not found??!?");
						#endif
						}

					DiscardSavedInsPt();
					break;

				case IDC_MAIN_DETAILS:
					if (recCurrent != noRecordSelected)
						{
						SaveInsPt();
						FrmPopupForm(FRM_DETAILS);
						}
					else
						Error(errNothingSelected);
					break;

				case IDC_CATEGORY_TRIGGER:
					// P11. If the category trigger is pressed, pop up the category list.
					SetCurrentRecord(noRecordSelected);
					DiscardSavedInsPt();
					MainViewSelectCategory();
					break;

				case IDC_STORE_TRIGGER:
					{
					Int ii;
					ListPtr plst;
					CharPtr rgpsz[MAX_STORES+2];

					DiscardSavedInsPt();
					plst = GetObjectPtr(IDC_STORE_LIST);
					for (ii = g_cStores; ii >= 0; ii--)
						rgpsz[ii] = g_rgStores[ii].pszName;
					IfGerman(rgpsz[g_cStores+1] = "Geschfte bearbeiten";)
					IfEnglish(rgpsz[g_cStores+1] = "Edit Stores";)
					LstSetListChoices(plst, rgpsz, 2+g_cStores);
					LstSetDrawFunction(plst, DrawStoreName);
					LstSetHeight(plst, 2+g_cStores);
					ii = LstPopupList(plst);
					if (ii == (Int)(g_cStores+1))
						FrmPopupForm(FRM_STORES);
					else if (ii != -1)
						{
						Assert(ii >= 0 && ii < (Int)(g_cStores+1));
						if (ii)
							g_store = g_rgStores[ii].bit;
						else
							// can't look at element 0, it's a list of all used store bits
							g_store = 0;
						Assert(!(g_store & storeOnce));
						SetStoreTrigger();
						irecTopVisible = firstRecord;
						ReloadTable(GetObjectPtr(IDC_TABLE));
						}
					}
					break;

				case IDC_VIEW_ALL:
					g_view = viewAll;
					goto reload_view;
				case IDC_VIEW_NEED:
					g_view = viewNeed;
					goto reload_view;
				case IDC_VIEW_COUPON:
					g_view = viewCoupon;
reload_view:
					SetCurrentRecord(noRecordSelected);
					DiscardSavedInsPt();
					irecTopVisible = firstRecord;
					ReloadTable(GetObjectPtr(IDC_TABLE));
					break;

				case IDC_PRICE:
					{
					ULong ul = 0;
					UInt record = firstRecord;
					VoidHand h;
					Item *pitem;
					Int offset = 0;

					TblReleaseFocus(GetObjectPtr(IDC_TABLE));

					// sum
					do
						{
						h = GetFilteredRecord(g_pdb, &record, catCurrent,
											  g_store, g_view, offset,
											  dmSeekForward);
						offset = 1;
						if (h)
							{
							pitem = MemHandleLock(h);
							ul += pitem->ulPrice * pitem->uQuantity;
							MemHandleUnlock(h);
							}
						}
					while (h);

					// display total
					FrmCustomAlert(ALT_MESSAGE,
								   IfGerman("Die Summe ist ")
								   IfEnglish("The total is ")
								   , LocalizeNumber(ul, fTrue, fFalse, 0), ".");

					RestoreInsPt();
					}
					break;

				case IDC_HELP:
					FrmHelp(STR_MAIN_INFO);
					RestoreInsPt();
					break;

				default:
					fHandled = fFalse;
				}
			break;

		case ctlRepeatEvent:
			// P13.  an onscreen scroll button is pressed
			switch (pEvent->data.ctlRepeat.controlID)
				{
				case IDC_UP:
					nDir = up;
					break;
				case IDC_DOWN:
					nDir = down;
					break;
				}
			if (nDir != -1)
				MainViewScroll(nDir);
			// Repeating controls don't repeat if handled is set true.
			fHandled = fFalse;
			break;

		case keyDownEvent:
			{
			FormPtr pfrm = FrmGetActiveForm();

			// P13.  Handle the physical scroll buttons
			switch (pEvent->data.keyDown.chr)
				{
				case pageUpChr:
					nDir = up;
					break;
				case pageDownChr:
					if (CtlEnabled(GetObjectPtr(IDC_DOWN)))
						nDir = down;
					break;
				}
			if (nDir != -1)
				MainViewScroll(nDir);
			else if (FrmGetFocus(pfrm) == FrmGetObjectIndex(pfrm, IDC_LOOKUP))
				{
				Char sz[cchMaxLookup+1];
				FieldPtr pfld = GetObjectPtr(IDC_LOOKUP);

				Assert(FldGetTextLength(pfld) < sizeof(sz));

				StrCopy(sz, _FldGetTextPtr(pfld));
				//$ review: (chrisant) normally FrmDispatchEvent would call
				// this; given how delicate PalmOS is, i am probably screwing
				// something up by doing this.
				fHandled = FldHandleEvent(pfld, pEvent);

				if (fHandled && StrCompare(sz, _FldGetTextPtr(pfld)) != 0)
					Lookup();
				}
			else
				fHandled = fFalse;
			}
			break;

		case tblRedrawFieldEvent:
			FldDrawField(TblGetCurrentField(pEvent->data.tblEnter.pTable));
			break;

		case tblDeleteAndReloadEvent:
			DeleteRecord(pEvent->data.tblEnter.row);	// yes, passing record in row member
			ReloadTable(pEvent->data.tblEnter.pTable);
			break;

		case tblSelectEvent:
			{
			Boolean f;
			Int row = pEvent->data.tblSelect.row;
			Int column = pEvent->data.tblSelect.column;
			FieldPtr pfld = GetObjectPtr(IDC_LOOKUP);

			FldFreeMemory(pfld);
			DrawBlankField(pfld);

			ptbl = pEvent->data.tblSelect.pTable;
			if (SetCurrentRecord(TblGetRowData(ptbl, row)))
				// if the current record got deleted, ignore this table event
				break;

			// P12. An entry in the table was selected.
			switch (column)
				{
				case COL_NEED:
				case COL_COUPON:
					f = ToggleItemFlag(ptbl, row, column);
					if (column == 0 && !f && g_view != viewAll)
						{
						FieldType fld;
						VoidHand h;
						Item *pitem;
						Int cy = TblGetRowHeight(ptbl, row);
						Int yy = FntLineHeight()/2;
						LineInfoPtr pli;
						FontID font;

						h = DmQueryRecord(g_pdb, TblGetRowData(ptbl, row));
						Assert(h);
						pitem = MemHandleLock(h);
						Assert(pitem);

						// use LineInfo in field to calc the cross-out line
						MemSet(&fld, sizeof(fld), 0);
						TblGetItemBounds(ptbl, row, COL_NAME, &fld.rect);
						FldSetText(&fld, h, sizeof(*pitem), pitem->cchName + 1);
						FldRecalculateField(&fld, fFalse);

						// draw line thru item
						font = FntSetFont(stdFont);
						for (pli = fld.lines; yy < cy;
							 yy += FntLineHeight(), pli++)
							{
							WinDrawLine(fld.rect.topLeft.x,
										fld.rect.topLeft.y + yy,
										fld.rect.topLeft.x + FntCharsWidth(fld.text + pli->start, pli->length) - 1,
										fld.rect.topLeft.y + yy);
							}
						FntSetFont(font);

						MemHandleUnlock(h);
						FldSetTextHandle(&fld, 0);
						FldFreeMemory(&fld);

						// wait 1/10th of second
						SysTaskDelay(sysTicksPerSecond/10);

						// delete record if it's an "only once" item
						if (GetItemStore(recCurrent) & storeOnce)
							DeleteRecord(recCurrent);

						SetCurrentRecord(noRecordSelected);
						DiscardSavedInsPt();
						ReloadTable(ptbl);
						}
					RestoreInsPt();
					break;

				case COL_QUANTITY:
					DoQuantity(ptbl, recCurrent, row, column);
					break;

				case COL_AISLE:
					DoAisle(ptbl, recCurrent, row, column);
					break;

				case COL_CATEGORY:
					{
					FormPtr		pfrm;
					TablePtr	ptbl;
					Word		catItem;
					Word		catNew;
					Boolean		fEdited;
					Char		sz[dmCategoryLength];
					UInt		attr;
					ListPtr		plst = GetObjectPtr(IDC_CATEGORY_LIST);
					RectangleType	rcBounds;
					RectangleType	rc;

					DiscardSavedInsPt();

					if (DmRecordInfo(g_pdb, recCurrent, &attr, NULL, NULL))
						break;

					catItem = (attr & dmRecAttrCategoryMask);
					catNew = catItem;

					pfrm = FrmGetActiveForm();
					ptbl = GetObjectPtr(IDC_TABLE);
					*sz = 0;	// don't care about this

					TblUnhighlightSelection(ptbl);

					rcBounds = plst->bounds;
					TblGetItemBounds(ptbl, row, column, &rc);
					LstSetPosition(plst, rc.topLeft.x, rc.topLeft.y);
					fEdited = CategorySelectV10(g_pdb, pfrm, IDC_DUMMY_TRIGGER,
												IDC_CATEGORY_LIST, fFalse, &catNew, sz);
					plst->bounds = rcBounds;

					if (fEdited || (catItem != catNew))
						{
						attr &= ~dmRecAttrCategoryMask;
						attr |= catNew;
						DmSetRecordInfo(g_pdb, recCurrent, &attr, NULL);
						if (catCurrent != dmAllCategories)
							{
							ReloadTable(ptbl);
							break;
							}
						}

					TblMarkRowInvalid(ptbl, row);
					TblRedrawTable(ptbl);
					}
					break;

				case COL_NAME:
					fHandled = fFalse;
					break;

				case COL_PRICE:
				default:
					Panic1("tblSelect: column %d unexpected", column);
				}
			}
			break;

		case tblEnterEvent:
			{
			Int row = pEvent->data.tblEnter.row;
			Int column = pEvent->data.tblEnter.column;

			ptbl = pEvent->data.tblSelect.pTable;
			switch (column)
				{
				case COL_PRICE:
					// if the current record got deleted, ignore this event
					if (!SetCurrentRecord(TblGetRowData(ptbl, row)))
						TableSelectPrice(pEvent, ptbl, row);
					break;

				default:
					TableSavePrice(ptbl);
					fHandled = fFalse;
				}
			}
			break;

		case fldHeightChangedEvent:
			{
			ptbl = GetObjectPtr(IDC_TABLE);
			Assert(recCurrent != noRecordSelected && TblEditing(ptbl));

			TblHandleEvent(ptbl, pEvent);
			IncrementalLoadTable(ptbl);
			TblRedrawTable(ptbl);
			MainViewUpdateScrollers(FrmGetActiveForm(),
									TblGetRowData(ptbl, g_cVisibleItems - 1));
			}
			break;

		case frmOpenEvent:
			// P3. The form was told to open.
			// P7. Initialize the main form.
			MainViewInit();
			break;

		case frmCloseEvent:
			TblReleaseFocus(GetObjectPtr(IDC_TABLE));	// just in case...
			SetCurrentRecord(noRecordSelected);
			fHandled = fFalse;
			break;

		case frmGotoEvent:
			{
			Word row;

			ptbl = GetObjectPtr(IDC_TABLE);

			SetCurrentRecord(pEvent->data.frmGoto.recordNum);
			MainViewInit();
			if (TblFindRowData(ptbl, recCurrent, &row))
				{
				TblSelectItem(ptbl, row, COL_NAME);
				}
			else
				{
				PanicSz("how can we get this far but not find the record??!?");
				Error(errErrorOpeningRecord);
				}
			}
			break;

		case fldChangedEvent:
			if (pEvent->data.fldChanged.fieldID == IDC_LOOKUP)
				Lookup();
			break;

		case menuEvent:
			// P2. A menu item was selected.
			ptbl = GetObjectPtr(IDC_TABLE);
			switch (pEvent->data.menu.itemID)
				{
				case IDM_ABOUT:
					FrmAlert(ALT_ABOUT);
					break;

				case IDM_DELETE:
					{
					if (recCurrent != noRecordSelected)
						{
						if (FrmCustomAlert(ALT_CONFIRM,
										   IfGerman("Diesen Artikel lschen?")
										   IfEnglish("Delete this item?")
										   , "", "") == 0)
							{
							TblReleaseFocus(ptbl);
							DeleteRecord(recCurrent);
							ReloadTable(ptbl);
							}
						}
					else
						Error(errNothingSelected);
					}
					break;

				case IDM_OPTIONS:
					FrmPopupForm(FRM_OPTIONS);
					break;

				case IDM_STORES:
					FrmPopupForm(FRM_STORES);
					break;

				case IDM_EXPORT:
					ExportToMemoPad();
					break;

				case IDM_MARK_ALL:
				case IDM_UNMARK_ALL:
					{
					Boolean fMark = (pEvent->data.menu.itemID == IDM_MARK_ALL);

					if (FrmCustomAlert(ALT_CONFIRM,
#if defined(GERMAN)
									   "Alle Artikel in diesem Geschft und in dieser Kategorie ",
									   fMark ? "als bentigt markieren?" : "auf nicht mehr bentigt zurcksetzen?",
									   ""
#else
									   fMark ? "Mark" : "Reset",
									   " all items in this store and category as ",
									   fMark ? "needed?" : "no longer needed?"
#endif
									   ) == 0)
						{
						UInt record = firstRecord;
						VoidHand h;
						Item *pitem;
						Int offset = 0;
						Boolean fDirty;

						TblReleaseFocus(ptbl);

						do
							{
							// use viewAll -- doesn't matter what the view is
							h = GetFilteredRecord(g_pdb, &record, catCurrent,
												  g_store, viewAll, offset,
												  dmSeekForward);
							offset = 1;
							if (h)
								{
								h = DmGetRecord(g_pdb, record);
								Assert(h);
								if (h)
									{
									pitem = MemHandleLock(h);
									fDirty = SetNeed(pitem, fMark);
									MemHandleUnlock(h);
									DmReleaseRecord(g_pdb, record, fDirty);
									}
								}
							}
						while (h);

						irecTopVisible = firstRecord;
						Assert(recCurrent == noRecordSelected);
						ReloadTable(ptbl);
						}
					}
					break;

				case IDM_SORT_BYNAME:
				case IDM_SORT_BYAISLE:
					{
					Boolean fAisle = pEvent->data.menu.itemID == IDM_SORT_BYAISLE;

					if (FrmCustomAlert(ALT_CONFIRM,
#if defined(GERMAN)
									   "Alle Artikel nach ",
									   fAisle ? "Gang" : "Name",
									   " sortieren?"
#else
									   "Sort all items by ",
									   fAisle ? "Aisle" : "Name",
									   "?"
#endif
									   ) == 0)
						{
						//$ review: alternatively could get uniqueID, sort,
						// find uniqueID and make it the current record again.
						SetCurrentRecord(noRecordSelected);

						DmQuickSort(g_pdb, CompareItemsByName, fAisle);
						ReloadTable(ptbl);
						}
					}
					break;

				case IDM_HOWTOUSE:
					FrmHelp(STR_HOWTOUSE);
					RestoreInsPt();
					break;

				default:
					fHandled = HandleEditMenu(pEvent);
					break;
				}
			break;

		default:
			fHandled = fFalse;
		}

	CALLBACK_EPILOGUE
	return fHandled;
}



///////////////////////////////////////////////////////////////////////////
// ApplicationHandleEvent

static Boolean ApplicationHandleEvent(EventPtr pEvent)
{
	FormPtr		pfrm;
	Int			idForm;
	Boolean		fHandled = fFalse;

	if (pEvent->eType == frmLoadEvent)
		{
		// Load the form resource specified in the event then activate the form.
		idForm = pEvent->data.frmLoad.formID;
		pfrm = FrmInitForm(idForm);
		Assert(pfrm);
		FrmSetActiveForm(pfrm);

		// Set the event handler for the form.  The handler of the currently 
		// active form is called by FrmDispatchEvent each time it receives an event.
		switch (idForm)
			{
			case FRM_MAIN:
				FrmSetEventHandler(pfrm, MainViewHandleEvent);
				break;

			case FRM_DETAILS:
				FrmSetEventHandler(pfrm, DetailsFormHandleEvent);
				break;

			case FRM_OPTIONS:
				FrmSetEventHandler(pfrm, OptionsFormHandleEvent);
				break;

			case FRM_STORES:
				FrmSetEventHandler(pfrm, StoresFormHandleEvent);
				break;

			case FRM_STORENAME:
				FrmSetEventHandler(pfrm, StoreNameFormHandleEvent);
				break;

			case FRM_QUANTITY:
				FrmSetEventHandler(pfrm, QuantityFormHandleEvent);
				break;

			default:
				Panic1("frmLoadEvent, idForm %d unknown", idForm);
			}
		fHandled = fTrue;
		}

	return fHandled;
}


static Boolean SpecialHandleEvent(EventPtr pEvent)
{
	Boolean fHandled = fFalse;
	RectangleType rc;

	switch (pEvent->eType)
		{
		case penDownEvent:
			// when Table is in edit mode, click on nothingness
			// should call TblReleaseFocus and set recCurrent ==
			// noRecordSelected.
			if (FrmGetActiveFormID() == FRM_MAIN)
				{
				FieldPtr pfld = GetObjectPtr(IDC_LOOKUP);

				FldGetBounds(pfld, &rc);
				if (!RctPtInRectangle(pEvent->screenX, pEvent->screenY, &rc))
					{
					// if focus is already in a field, the field says it
					// didn't handle the penDownEvent.  so we must explicitly
					// check if the pen was in the field.  sheesh.
					FldDelete(pfld, 0, FldGetTextLength(pfld));
					FldReleaseFocus(pfld);
					}

				if (recCurrent != noRecordSelected)
					{
					TblReleaseFocus(GetObjectPtr(IDC_TABLE));
					SetCurrentRecord(noRecordSelected);
					DiscardSavedInsPt();
					fHandled = fTrue;
					}
				}
			break;

		default:
			break;
		}
	return fHandled;
}



///////////////////////////////////////////////////////////////////////////
// EventLoop

static void EventLoop(void)
{
	EventType	event;
	Word		error;
	
	do
		{
		// Get the next available event.
		EvtGetEvent(&event, evtWaitForever);

		// Give the system a chance to handle the event.
		if (!SysHandleEvent(&event))

			// P2. Give the menu bar a chance to update and handle the event.
			if (!MenuHandleEvent(0, &event, &error))

				// P3. Give the application a chance to handle the event.
				if (!ApplicationHandleEvent(&event))

					// P3. Let the form object provide default handling of the event.
					if (!FrmDispatchEvent(&event))
						SpecialHandleEvent(&event);
		}
	while (event.eType != appStopEvent);
}



///////////////////////////////////////////////////////////////////////////
// PilotMain

DWord PilotMain(Word cmd, Ptr pcmdPBP, Word launchFlags)
{
	CALLBACK_PROLOGUE

	switch (cmd)
		{
		case sysAppLaunchCmdNormalLaunch:
			if (StartApplication())
				{
				FrmGotoForm(FRM_MAIN);
				EventLoop();
				StopApplication();
				}
			break;

		case sysAppLaunchCmdFind:
			Search((FindParamsPtr)pcmdPBP);
			break;

		case sysAppLaunchCmdGoTo:
			// Check if new global variables are set up.  This occurs when the
			// system intends for the app to start.
			if (launchFlags & sysAppLaunchFlagNewGlobals)
				{
				// This occurs when the user wants to see a record in this
				// application's database.  The app must be started and
				// a view to display the data must be opened.
				if (StartApplication())
					{
					GoToRecord((GoToParamsPtr)pcmdPBP, fTrue);
					EventLoop();
					StopApplication();
					}
				}
			else
				{
				// The app was already running.  Open a form to display the
				// record.
				GoToRecord((GoToParamsPtr)pcmdPBP, fFalse);
				}
			break;

		case sysAppLaunchCmdSaveData:
			// P14. Launch code sent to running app to save all data.
			// This is neccessary before a find for some applications because
			// some apps do not edit in place and the user might search for
			// data not saved.  Or saving a record might reorder the database.
			// Saving before a find avoids finding a record, saving,
			// resorting, and then displaying a record moved to the position
			// of the found record.
			FrmSaveAllForms();
			break;

		case sysAppLaunchCmdSyncNotify:
			DbgMsg((20, "sysAppLaunchCmdSyncNotify"));
			// P17. The database has been synchronized.  Handle any changes
			// in the data (resorting data could be performed).
			//SyncNotification();
			break;

		case sysAppLaunchCmdInitDatabase:
			DbgMsg((20, "sysAppLaunchCmdInitDatabase"));
			// P17. The database has been created but there isn't an app info
			// block. Either the HotSync app or another application which uses
			// this application's data may request that the application info
			// block be created by this app (because it's the creator of the
			// database).
			//$ review: (chrisant) why in the world is this getting called we
			// installing a perfectly complete backed up database that has an
			// AppInfo block?  even though this supposedly wipes it, the
			// installed database's AppInfo block is intact, though.
			InitAppInfo(((SysAppLaunchCmdInitDatabaseType*)pcmdPBP)->dbP, fFalse);
			DbgMsg((20, "sysAppLaunchCmdInitDatabase SUCCEEDED"));
			break;

		default:
			DbgMsg((20, "PilotMain, cmd == %d", cmd));
			break;
		}

	CALLBACK_EPILOGUE
	return 0;
}

