///////////////////////////////////////////////////////////////////////////
// stores.c
//
// the Edit Stores form

#ifndef _MAIN_H
#include "main.h"
#endif

#ifndef _STORES_H
#include "stores.h"
#endif

#ifndef _UTIL_H
#include "util.h"
#endif

#include "res.h"



///////////////////////////////////////////////////////////////////////////
// Stores Form

static Boolean s_fDirtyStoresForm;
static Int s_iStore;

static Char s_szStoreName[cchMaxStoreName+1];
static CharPtr s_rgpszStores[MAX_STORES];


Err ZapStore(DmOpenRef pdb, UInt bit)
{
	Err			err = 0;
	UInt		record;
	VoidHand	h;
	Item		*pitem;
	Boolean		fMatch;
	UInt		store;
#ifdef DEBUG
	ULong		ticks = TimGetTicks();
	UInt		bitOriginal = bit;
#endif

	Assert(pdb);
	Assert(bit);

	// loop thru all records in database, clear the specified store bit
	for (record = DmNumRecords(pdb); record--;)
		{
		h = DmQueryRecord(pdb, record);
		if (h)
			{
			pitem = MemHandleLock(h);

			// didn't need local var for this, but makes code cleaner (more clear
			// exactly what will happen in an error condition)!!
			fMatch = (pitem->store & bit);
			MemHandleUnlock(h);

#if 0
fMatch = fTrue;
#endif
			if (fMatch)
				{
				h = DmGetRecord(pdb, record);
				if (!h)
					{
					Panic1("record %d probably busy.", record);
					goto GetLastError;
					}

#if 0
{
UChar cchName = pitem->cchName;
AssertSameSize(cchName, pitem->cchName);
DmWrite(pitem, OffsetOf(*pitem, cchName), &cchName, sizeof(pitem->cchName));
}
#endif
				// clear the bit
				pitem = MemHandleLock(h);
				store = (pitem->store & (~bit));
				Assert(sizeof(pitem->store) == sizeof(store));
				DmWrite(pitem, OffsetOf(*pitem, store), &store, sizeof(pitem->store));
				MemHandleUnlock(h);
				DmReleaseRecord(pdb, record, fTrue);
				}
			}
		}

	goto Error;
GetLastError:
	err = DmGetLastErr();
Error:
	DbgMsg((20, "ZapStore took %d clock ticks", TimGetTicks()-ticks));
	Assert1(!err, "oops, error = %d", err);
	Assert(bit == bitOriginal);
	return err;
}


static Int CompareStores(VoidPtr pA, VoidPtr pB, Long other)
{
	return StrCompare(((Store*)pA)->pszName, ((Store*)pB)->pszName);
}


static void SortStores(void)
{
	SysInsertionSort(&g_rgStores[1], g_cStores, sizeof(g_rgStores[1]),
					 CompareStores, 0);
}


static void InitForm(void)
{
	Int			ii;

	s_fDirtyStoresForm = fFalse;

	for (ii = g_cStores; ii--;)
		s_rgpszStores[ii] = g_rgStores[ii+1].pszName;
	LstSetListChoices(GetObjectPtr(IDC_STORE_LIST), s_rgpszStores, g_cStores);

	FrmDrawForm(FrmGetActiveForm());
}


Boolean StoresFormHandleEvent(EventPtr pEvent)
{
	Boolean		fHandled = fTrue;
	Int			ii;

	CALLBACK_PROLOGUE

	switch (pEvent->eType)
		{
		case ctlSelectEvent:
			// A control button was pressed and released.
			switch (pEvent->data.ctlSelect.controlID)
				{
				case IDC_FORM_OK:
					if (s_fDirtyStoresForm)
						{
						FrmCloseAllForms();
						FrmGotoForm(FRM_MAIN);
						}
					else
						{
						FrmReturnToForm(0);
						RestoreInsPt();
						}
					break;

				case IDC_STORES_NEW:
					if (g_cStores < MAX_STORES)
						{
						s_szStoreName[0] = 0;
						s_iStore = -1;
						FrmPopupForm(FRM_STORENAME);
						}
					else
						Error(errTooManyStores);
					break;

				case IDC_STORES_RENAME:
					{
					ii = LstGetSelection(GetObjectPtr(IDC_STORE_LIST));
					if (ii != -1)
						{
						s_iStore = ii+1;
						Assert(g_rgStores[s_iStore].pszName);

						Assert(StrLen(g_rgStores[s_iStore].pszName) < cchMaxStoreName);
						StrCopy(s_szStoreName, g_rgStores[s_iStore].pszName);
						FrmPopupForm(FRM_STORENAME);
						}
					}
					break;

				case IDC_STORES_DELETE:
					ii = LstGetSelection(GetObjectPtr(IDC_STORE_LIST));
					if (ii != -1)
						{
						// the items in the store, just the store itself.
						if (FrmCustomAlert(ALT_CONFIRM,
#if defined(GERMAN)
										   "Geschft \0x84",
										   g_rgStores[ii+1].pszName,
										   "\0x94 lschen (Die Artikel im Laden werden nicht gelscht)?"
#else
										   "Delete store \"",
										   g_rgStores[ii+1].pszName,
										   "\"?  (This does not delete the items in the store)"
#endif
										   ) == 0)
							{
							UInt bit = g_rgStores[ii+1].bit;

							g_cStores--;
							MemPtrFree(g_rgStores[ii+1].pszName);

							// remove the bit from the list of used bits
							g_rgStores[0].bit &= ~bit;
							g_store = 0;

							// clear this store bit from all records
							ZapStore(g_pdb, bit);

							// shift other store info
							if (ii < (Int)g_cStores)
								MemMove(g_rgStores+ii+1, g_rgStores+ii+2,
										sizeof(g_rgStores[0])*(g_cStores-ii));

							InitForm();		// must do this BEFORE setting s_fDirtyStoresForm, since it clears s_fDirtyStoresForm
							s_fDirtyStoresForm = fTrue;
							}
						}
					break;

				default:
					fHandled = fFalse;
					break;
				}
			break;

		case frmOpenEvent:
			InitForm();
			break;

		case frmCloseEvent:
			// this does NOT get sent by FrmReturnToForm.  must delete the
			// form when closed other than by FrmReturnToForm.
			FrmDeleteForm(FrmGetActiveForm());
			break;

		default:
			fHandled = fFalse;
		}

	CALLBACK_EPILOGUE
	return fHandled;
}


Boolean StoreNameFormHandleEvent(EventPtr pEvent)
{
	Boolean fHandled = fTrue;

	switch (pEvent->eType)
		{
		case ctlSelectEvent:
			// A control button was pressed and released.
			switch (pEvent->data.ctlSelect.controlID)
				{
				case IDC_FORM_OK:
					{
					CharPtr psz = 0;
					CharPtr pszField;
					UInt bit;
					FieldPtr pfld = GetObjectPtr(IDC_STORE_NAME_FIELD);

					if (FldDirty(pfld))
						{
						pszField = FldGetTextPtr(pfld);
						if (pszField && *pszField)
							psz = MemPtrNew(FldGetTextLength(pfld)+1 _MEMTRACK);
						if (psz)
							{
							if (s_iStore == -1)
								{
								g_cStores++;
								s_iStore = g_cStores;

								// allocate new bit
								for (bit = 0x0001; bit && (g_rgStores[0].bit & bit); bit <<= 1);
								Assert(bit);
								Assert(bit != storeOnce);
								g_rgStores[0].bit |= bit;
								g_rgStores[s_iStore].bit = bit;
								g_rgStores[s_iStore].cItems = 0;
								}
							else
								// renaming, so free current string
								MemPtrFree(g_rgStores[s_iStore].pszName);

							g_rgStores[s_iStore].pszName = psz;
							StrCopy(psz, pszField);

							SortStores();

							// mark the Store Form dirty (ie, the form that
							// launched this Store Name form).
							FrmReturnToForm(0);
							InitForm();		// must do this BEFORE setting s_fDirtyStoresForm, since it clears s_fDirtyStoresForm
							s_fDirtyStoresForm = fTrue;
							}
						break;
						}
					}
					// else fall thru
				case IDC_FORM_CANCEL:
					FrmReturnToForm(0);
					break;

				default:
					fHandled = fFalse;
				}
			break;

		case menuEvent:
			fHandled = HandleEditMenu(pEvent);
			break;

		case frmOpenEvent:
			{
			FormPtr		pfrm = FrmGetActiveForm();
			FieldPtr	pfld = GetObjectPtr(IDC_STORE_NAME_FIELD);
			Int			cch = StrLen(s_szStoreName);

			if (cch)
				{
				Verify(FldInsert(pfld, s_szStoreName, cch));

				//$ review: (chrisant) this doesn't work as advertised
				//FldSetSelection(pfld, 0, cch);
				FldSetInsPtPosition(pfld, cch);
				}
			FrmDrawForm(pfrm);

			FrmSetFocus(pfrm, FrmGetObjectIndex(pfrm, IDC_STORE_NAME_FIELD));
			}
			break;

		case frmCloseEvent:
			// this does NOT get sent by FrmReturnToForm.  must delete the
			// form when closed other than by FrmReturnToForm.
			FrmDeleteForm(FrmGetActiveForm());
			break;

		default:
			fHandled = fFalse;
		}
	return fHandled;
}



