///////////////////////////////////////////////////////////////////////////
// util.c
//
// utility functions


#ifndef _MAIN_H
#include "main.h"
#endif

#include "res.h"



static Word s_wInsPtPos;
UInt s_record = noRecordSelected;
static Word s_column;
static Word s_wScrollPos;
static Word s_selStart;
static Word s_selEnd;



void Error(UInt idsErr)
{
	CharPtr psz = MemPtrNew(128 _MEMTRACK);

	SysCopyStringResource(psz, idsErr);

	DbgMsg((20, psz));

	FrmCustomAlert(ALT_ERRORMESSAGE, psz, 0, 0);
	MemPtrFree(psz);
}


VoidPtr GetObjectPtr(Int objectID)
{
	FormPtr		pfrm;

	pfrm = FrmGetActiveForm();
	return FrmGetObjectPtr(pfrm, FrmGetObjectIndex(pfrm, objectID));
}


FieldPtr GetFocusObjectPtr(void)
{
	FormPtr		pfrm;
	UInt		focus;

	// Get a pointer to tha active form and the index of the form object with focus.
	pfrm = FrmGetActiveForm();
	Assert(pfrm);
	focus = FrmGetFocus(pfrm);
	Assert(focus);

	// If no object has the focus return NULL pointer.
	if (focus == noFocus)
		return NULL;

	// Return a pointer to the object with focus.
	return FrmGetObjectPtr(pfrm, focus);
}


CharPtr _FldGetTextPtr(FieldPtr pfld)
{
	CharPtr psz = FldGetTextPtr(pfld);

	if (!psz)
		psz = "";
	return psz;
}


void InvertEditIndicator(TablePtr ptbl, Word row)
{
	Word column;
	RectangleType rc;

	if (row == (Word)-1)
		TblGetSelection(ptbl, &row, &column);
	TblGetItemBounds(ptbl, row, COL_NEED, &rc);
	WinInvertRectangle(&rc, 3);
}


void DrawBlankField(FieldPtr pfld)
{
	RectangleType rc;

	// good grief.  drawing a field grabs the focus.  how stupid.  so if i
	// want to redraw a field without grabbing focus (say, in response to
	// tblSelectEvent) i have to draw it myself.  lucky for me i only want to
	// draw a blank field (immediately after calling FldFreeMemory).

	FldGetBounds(pfld, &rc);
	WinEraseRectangle(&rc, 0);
	//$ review: (chrisant) why in blue blazes do i have to subtract 2 from the y coordinate??
	WinDrawGrayLine(rc.topLeft.x, rc.topLeft.y+rc.extent.y-2,
					rc.topLeft.x+rc.extent.x-1, rc.topLeft.y+rc.extent.y-2);
}


void DrawHelpButton(FormPtr pfrm, UInt idc)
{
	RectangleType rc;
	FontID font;
	Char ch;

	FrmGetObjectBounds(pfrm, FrmGetObjectIndex(pfrm, idc), &rc);
	WinEraseRectangle(&rc, 4);
	font = FntSetFont(symbolFont);
	ch = symbolHelp;
	WinDrawChars(&ch, 1, rc.topLeft.x + 3, 2);
	FntSetFont(font);
}


UInt TruncateLabel(CharPtr psz, Int cxMax)
{
	SWord			cxOrig = cxMax;
	SWord			cx;
	SWord			cchPrev;
	SWord			cch;
	Boolean			fTruncated;

	cch = StrLen(psz);
	while (fTrue)
		{
		cx = cxOrig;
		cchPrev = cch;
		//$ review: (chrisant) fTruncated does not seem to match the docs
		FntCharsInWidth(psz, &cx, &cch, &fTruncated);
		if (cch == cchPrev /*|| !fTruncated*/)
			break;
		if (cch)
			psz[cch-1] = horizEllipsisChr;
		psz[cch] = 0;
		if (!cch)
			break;
		}
	return cch;
}


Int CalcRowHeight(CharPtr psz, Int cx, Int cyLinesAvailable)
{
	Int cLines;
	FontID font;

	font = FntSetFont(stdFont);
	cLines = FldCalcFieldHeight(psz, cx);
	FntSetFont(font);
	return min(cLines, cyLinesAvailable);
}


Int GetStoreIndex(UInt store)
{
	Int ii;

	// the for loop is deceptively complex, be very careful when modifying it!
	for (ii = g_cStores; ii && store != g_rgStores[ii].bit; ii--);
	return ii;
}


void SetStoreTrigger(void)
{
	FontID			font;
	static Char		szLabel[cchMaxStoreName+1];

	StrCopy(szLabel, g_rgStores[GetStoreIndex(g_store)].pszName);

	font = FntSetFont(boldFont);
	TruncateLabel(szLabel, g_cxMaxStorePixels);
	FntSetFont(font);

	//$ todo: (chrisant) note, this holds on to the pointer but i think does not free the pointer?  true for all labels?
	CtlSetLabel(GetObjectPtr(IDC_STORE_TRIGGER), szLabel);
}


Word GetItemStore(UInt record)
{
	VoidHand	h;
	Item		*pitem;
	Word		store;

	h = DmQueryRecord(g_pdb, record);
	pitem = MemHandleLock(h);
	store = pitem->store;
	MemHandleUnlock(h);

	return store;
}


void UpdateCounts(Int delta, UInt bits)
{
	UInt bit;

	Assert(delta == 1 || delta == -1);

	g_rgStores[0].cItems += delta;
	for (bit = 0x4000; bit; bit >>= 1)
		if (bits & bit)
			g_rgStores[GetStoreIndex(bit)].cItems += delta;
}


void DeleteRecord(UInt record)
{
	VoidHand		h;
	Item			*pitem;

	h = DmQueryRecord(g_pdb, record);
	Assert(h);
	if (h)
		{
		pitem = MemHandleLock(h);
		if (pitem->flags & flagNeed)
			UpdateCounts(-1, pitem->store);
		MemHandleUnlock(h);
		}

	DmDeleteRecord(g_pdb, record);
	if (record == recCurrent)
		recCurrent = noRecordSelected;
}


Boolean FCanDelete(TablePtr ptbl, UInt record)
{
	VoidHand	h;
	ItemPtr		pitem;
	Boolean		fCanDelete = fFalse;

	if (ptbl && TblEditing(ptbl) && ptbl->currentColumn == COL_NAME &&
		TblGetRowData(ptbl, ptbl->currentRow) == record)
		{
		fCanDelete = !FldGetTextLength(TblGetCurrentField(ptbl));
		}
	else
		{
		h = DmQueryRecord(g_pdb, record);
		if (h)
			{
			pitem = MemHandleLock(h);
			fCanDelete = (pitem->cchName == 0);
			Assert(pitem->cchName == StrLen(pitem->szName));
			MemHandleUnlock(h);
			}
		}
	return fCanDelete;
}


Boolean SetCurrentRecord(UInt record)
{
	Boolean fCanDelete = fFalse;
	TablePtr ptbl = GetObjectPtr(IDC_TABLE);

	Assert(ptbl);

	if (recCurrent != record)
		{
		TableSavePrice(ptbl);

		if (recCurrent != noRecordSelected)
			{
			// done with current record
			fCanDelete = FCanDelete(ptbl, recCurrent);
			if (fCanDelete)
				{
				// the Name field is empty, so delete the record
				SendDeleteAndReloadTable(ptbl, recCurrent);
				}
			}

		recCurrent = record;
		}
	return fCanDelete;
}


#if 0
void SendPenDownEvent(PointType *ppt)
{
	EventType event;

	event.eType = penDownEvent;
	event.penDown = fTrue;
	event.screenX = ppt->x;
	event.screenY = ppt->y;
	EvtAddEventToQueue(&event);
}
#endif


void SendUpdateTableFieldRedraw(TablePtr ptbl)
{
	EventType event;

	event.eType = tblRedrawFieldEvent;
	event.data.tblEnter.pTable = ptbl;
	EvtAddEventToQueue(&event);
}


void SendDeleteAndReloadTable(TablePtr ptbl, UInt record)
{
	EventType event;

	event.eType = tblDeleteAndReloadEvent;
	event.data.tblEnter.pTable = ptbl;
	event.data.tblEnter.row = record;
	EvtAddEventToQueue(&event);
}


CharPtr LocalizeNumber(ULong ul, Boolean fCurrencySymbol, Boolean fForceDecimals, UInt *pcch)
{
	static Char sz[20];
	CharPtr psz = &sz[sizeof(sz)-1];
	Int ii;

	*psz = 0;

	if (ul || fCurrencySymbol)
		{
		if (!g_fDecimals && fForceDecimals && (ul % 100) == 0)
			{
			// number is a whole number and rounding is on, so don't show any
			// decimal places.
			ul /= 100;
			}
		else if (g_fDecimals || fForceDecimals)
			{
			// two decimal places
			for (ii = 2; ii--; ul /= 10)
				*(--psz) = (ul % 10) + '0';
			*(--psz) = g_chDecimal;
			}
		else
			{
			//$ review: (chrisant) do stores always round up from .5?
			if ((ul % 100) >= 50)
				ul += 100;
			ul /= 100;
			}

		// whole number
		do
			{
			*(--psz) = (ul % 10) + '0';
			ul /= 10;
			}
		while (ul);

		// currency symbol
		if (fCurrencySymbol)
			{
			IfGerman(*(--psz) = ' '; *(--psz) = 'M'; *(--psz) = 'D';)
			IfEnglish(*(--psz) = '$';)
			}
		}
	if (pcch)
		*pcch = (sz + (sizeof(sz) - 1) - psz);

	return psz;
}


VoidHand GetFilteredRecord(DmOpenRef pdb, UInt *piRecord, UInt category,
						   UInt store, UInt view, Int cRecords, Int direction)
{
	Err				err;
	VoidHand		h = 0;
	Item			*pitem;
	Boolean			fFound = fFalse;
	UInt			recordPrev;
	UInt			record = *piRecord;
	Int				offset = (cRecords == 0) ? 0 : 1;

	cRecords = Abs(cRecords);

	do
		{
		h = 0;
		recordPrev = record;
		CE(DmSeekRecordInCategory(pdb, &record, offset, direction, category));
		if (record == recordPrev && offset != 0)
			goto Error;

		h = DmQueryRecord(g_pdb, record);
		if (h)
			{
			pitem = MemHandleLock(h);
			Assert(pitem);

			// filter unwanted items
			if ((view != viewAll && !(pitem->flags & flagNeed)) ||
				(view == viewCoupon && !(pitem->flags & flagCoupon)) ||
				(store && !(store & pitem->store)) ||
				(g_pSysPrefs->hideSecretRecords && (pitem->flags & flagPrivate)))
				{
				// not found
				offset = 1;
				}
			else
				{
				if (cRecords)
					cRecords--;
				if (cRecords == 0)
					{
					fFound = fTrue;
					*piRecord = record;
					}
				}

			MemHandleUnlock(h);
			}
		}
	while (h && !fFound);

Error:
	return h;
}


void ExportToMemoPad(void)
{
	FieldType	fld;
	Int			iStore = GetStoreIndex(g_store);
	UInt		record = firstRecord;
	VoidHand	h = 0;
	VoidHand	hNew;
	Item		*pitem;
	CharPtr		psz;
	VoidHand	hText;
	Int			offset = 0;
	DmOpenRef	pdbMemoPad;
	Word		cch;
	IfDebug(Boolean fVerify);

	// initialize the field
	MemSet(&fld, sizeof(fld), 0);
	fld.attr.singleLine = fTrue;
	fld.attr.editable = fTrue;
	fld.maxChars = 4000;			//$ review: (chrisant) limit to 4000 characters

	// open MemoPad database
	pdbMemoPad = DmOpenDatabaseByTypeCreator(memoPadDBType, memoPadAppType,
											 dmModeReadWrite);
	if (!pdbMemoPad)
		{
		Error(errNoMemoPad);
		goto Error;
		}

	//$ review: (chrisant) what if MemoPad database does not have an AppInfo block?

	//$ review: (chrisant) i think this could be more efficient if i created
	// the new record and directly put its chunk handle into the field, but
	// this way is safer, i think.  i can try fancy stuff later.
	hText = MemHandleNew(255);
	psz = MemHandleLock(hText);
	MemSet(psz, 255, 0);
//	*psz = 0;
	MemHandleUnlock(hText);
	FldSetText(&fld, hText, 0, 255);

	//if (!FldInsert(&fld, "Shopping List (<Need>):\n\t[store]\n\t[category]"))
	if (!FldInsert(&fld, g_rgStores[iStore].pszName, StrLen(g_rgStores[iStore].pszName)))
		{
		PanicSz("error inserting store name");
		goto Error;
		}
	if (!FldInsert(&fld,
				   IfGerman(" Einkaufsliste:")
				   IfEnglish(" Shopping List:")
				   , 15))
		{
		PanicSz("error inserting Shopping List");
		goto Error;
		}

	while (fTrue)
		{
		h = GetFilteredRecord(g_pdb, &record, catCurrent,
							  g_store, g_view, offset, dmSeekForward);
		if (!h)
			break;

		offset = 1;
		pitem = MemHandleLock(h);

		if (!FldInsert(&fld, "\n ", 3))
			{
			PanicSz("error inserting bullet");
			goto Error;
			}
		if (!FldInsert(&fld, pitem->szName, pitem->cchName))
			{
			PanicSz("error inserting item name");
			goto Error;
			}

		MemHandleUnlock(h);
		}

	cch = FldGetTextLength(&fld);
	record = 0;		// at index 0 in MemoPad database
	hNew = DmNewRecord(pdbMemoPad, &record, cch+1);
	if (!hNew)
		{
		Error(errErrorExporting);
		goto Error;
		}

	// copy text to new MemoPad record
	IfDebug(fVerify = !)
	DmWrite(MemHandleLock(hNew), 0, MemHandleLock(hText), cch+1);
	IfDebug(Assert1(fVerify, "DmWrite failed in ExportToMemoPad, error = %d.",
					DmGetLastErr());)
	MemHandleUnlock(hText);
	MemHandleUnlock(hNew);
	DmReleaseRecord(pdbMemoPad, record, fTrue);
	DmCloseDatabase(pdbMemoPad);

Error:
	if (h)
		MemHandleUnlock(h);
	FldFreeMemory(&fld);
	Assert(fld.text == 0);
	Assert(fld.textHandle == 0);
	Assert(fld.lines == 0);
}


void SaveInsPt(void)
{
	TablePtr ptbl;
	FieldPtr pfld;

	if (recCurrent != noRecordSelected)
		{
		ptbl = GetObjectPtr(IDC_TABLE);
		if (TblEditing(ptbl))
			{
			pfld = TblGetCurrentField(ptbl);
			Assert(pfld);
			if (pfld)
				{
				s_wInsPtPos = FldGetInsPtPosition(pfld);
				s_record = recCurrent;
				s_column = ptbl->currentColumn;
				FldGetSelection(pfld, &s_selStart, &s_selEnd);
				s_wScrollPos = FldGetScrollPosition(pfld);
				}
			}
		}
}


void DiscardSavedInsPt(void)
{
	s_record = noRecordSelected;
}


void RestoreInsPt(void)
{
	TablePtr		ptbl;
	FormPtr			pfrm;
	Word			row;
	FieldPtr		pfld;
	Boolean			fPrice = (s_column == COL_PRICE);
	RectangleType	rc;

	Assert(FrmGetActiveFormID() == FRM_MAIN);

	if (s_record != noRecordSelected)
		{
		ptbl = GetObjectPtr(IDC_TABLE);
		pfrm = FrmGetActiveForm();

		if (TblFindRowData(ptbl, s_record, &row))
			{
			FrmSetFocus(pfrm, FrmGetObjectIndex(pfrm, ptbl->id));
			if (fPrice)
				{
				ptbl->attr.editing = fTrue;
				ptbl->currentRow = row;
				ptbl->currentColumn = COL_PRICE;
				}
			else
				TblGrabFocus(ptbl, row, s_column);

			pfld = TblGetCurrentField(ptbl);
			Assert(pfld);

			if (pfld)
				{
				if (fPrice)
					{
					g_fEditingPrice = fTrue;
					TblGetItemBounds(ptbl, row, COL_PRICE, &rc);
					TableInitPrice(ptbl, pfld, row, COL_PRICE, &rc, fTrue);

					InvertEditIndicator(ptbl, row);
					}

				FldSetScrollPosition(pfld, s_wScrollPos);
				FldSetInsPtPosition(pfld, s_wInsPtPos);
				FldSetSelection(pfld, s_selStart, s_selEnd);
				FldGrabFocus(pfld);
				}
			}
		DiscardSavedInsPt();
		}
}


Boolean HandleEditMenu(EventPtr pEvent)
{
	Boolean fHandled = fTrue;
	FieldPtr pfld = GetFocusObjectPtr();
	TablePtr ptbl = 0;

	if (FrmGetActiveFormID() == FRM_MAIN)
		ptbl = GetObjectPtr(IDC_TABLE);

	// handle fields inside table
	if (ptbl && TblEditing(ptbl))
		pfld = TblGetCurrentField(ptbl);

	Assert(pEvent->eType == menuEvent);

	switch (pEvent->data.menu.itemID)
		{
		case IDM_CUT:			// Do a cut to the clipboard if text is highlighted.
			if (pfld)
				FldCut(pfld);
			break;

		case IDM_COPY:			// Copy the highlighted text to the clipboard.
			if (pfld)
				FldCopy(pfld);
			break;

		case IDM_PASTE:			// Do a paste from the clipboard.
			if (pfld)
				FldPaste(pfld);
			break;
	
		case IDM_UNDO:			// Undo the last text change.
			if (pfld)
				FldUndo(pfld);
			break;

		case IDM_SELECTALL:		// Select all the existing text.
			if (pfld)
				FldSetSelection(pfld, 0, FldGetTextLength(pfld));
			break;
	
		case IDM_KEYBOARD:		// Display the on screen keyboard.
			if (pfld)
				{
#if 0
				FieldAttrType attr;

				FldGetAttributes(pfld, &attr);
				SysKeyboardDialog(attr.numeric ? kbdNumbersAndPunc : kbdAlpha);
#endif
				SysKeyboardDialog(kbdDefault);
				}
			break;

		case IDM_GRAFFITI:
			SysGraffitiReferenceDialog(referenceAlpha);
			break;

		default:
			fHandled = fFalse;
		}
	return fHandled;
}

